# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import enum
import plistlib
from cryptography.hazmat.primitives.ciphers import algorithms, modes, Cipher

try:
    from rich import print
except ImportError:
    pass
# ---------------------------------------------------------------------
# Key information for the structs were taken from here:
#   - https://home.vollink.com/gary/playlister/ituneslib.html
#   - https://gitlab.home.vollink.com/external/musicdb-poc/
#   - Internal TypeLib of iTunes (accessible from within the application
#     binary)
# ---------------------------------------------------------------------

from caterpillar.model import struct
from caterpillar.shortcuts import *
from caterpillar.fields import *

opt.set_struct_flags(opt.S_DISCARD_UNNAMED)

KEY = None


# Unfortunately, we have to implement the decryption process on
# our own. But let it be an example of how to extend this library.
@singleton
class EncryptedBlob(Transformer):
    # The base struct will be a memory instance
    def __init__(self) -> None:
        super().__init__(Memory(...))

    # PACKING:
    def encode(self, obj: bytes, context) -> bytes:
        # we just have to encrypt the content before it gets written
        # to the stream.
        cipher = Cipher(algorithms.AES(KEY), modes.ECB())
        encryptor = cipher.encryptor()

        # Encryption is a little bit different. The last 16 bytes are
        # static and won't be encrypted
        length = len(obj)
        pos = length - (length % 16)
        encrypted = encryptor.update(obj[:pos]) + encryptor.finalize()
        # usually, an instance of Encrypted(...) would be used here
        return encrypted + obj[pos:]

    # UNPACKING
    def decode(self, parsed: memoryview, context) -> bytes:
        # the same again: note that the parsed object is encrypted
        # and must be decompressed afterwards
        cipher = Cipher(algorithms.AES(KEY), modes.ECB())
        decryptor = cipher.decryptor()

        # REVISIT: Trailer struct that servers as a copy-only padding
        length = len(parsed)
        pos = length - (length % 16)
        decrypted = decryptor.update(parsed[:pos]) + decryptor.finalize()
        return decrypted + parsed[pos:]


# We create a chain to first, decrypt the data, then decompress it
# and later on, we can use this chain to parse chunks from the
# decompressed data.
ITDBEncrypted = EncryptedBlob & ZLibCompressed(..., comp_kwargs={"level": 1})


# This will be the base class for all data chunks
@struct(order=LittleEndian)
class MHChunk:
    type: Bytes(4)
    header_len: uint32
    total_len: uint32


# First of all, define the global header
@struct(order=LittleEndian)
class MHFileData(MHChunk):  # mdfh
    unknown: uint32
    # Prefixed wouldn't help here, as the string is fixed. We could add
    # a padding to the prefixed class.
    version: Prefixed(uint8, encoding="utf-8")
    _: padding[31 - lenof(this.version)]
    chunk_count: uint32
    # This id is important to iTunes
    library_id: uint64
    # the rest of the data will be copied
    pad: Memory(this.header_len - 60)


class ObjectType(enum.IntEnum):
    __struct__ = uint32

    kTrackName = 0x02
    kTrackAlbumName = 0x03
    kTrackArtist = 0x04
    kTrackGenre = 0x05
    kTrackKind = 0x06
    kTrackEQ = 0x07
    kTrackComments = 0x08
    kTrackCategory = 0x09
    kTrackFileURL = 0x0B
    kTrackComponist = 0x0C
    kTrackLocation = 0x0D
    kTrackGroup = 0x0E

    kTrackShortDescription = 0x12
    kTrackFullDescription = 0x16
    kTVShowName = 0x18
    kTVEpisode = 0x19
    kTrackAlbumArtist = 0x1B

    kSmartPlaylistTitle = 0x64
    kSmartCriteria = 0x65
    kSmartInfo = 0x66
    kSmartPList = 0x6D

    kAlbumName = 0x12C
    kAlbumArtist = 0x12D
    kAlbumArtistSort = 0x12E

    kVideoDirector = 0x190
    kVideoDirectorSort = 0x191

    kPlaySmartFilter = 0x2BC
    kPLaySmartName = 0x2BE
    kPlaySmartInfo = 0x2BF

    kPlist = 0x320


class StringType(enum.IntEnum):
    __struct__ = uint32

    kURI = 0
    kWString = 1
    kEscapedURI = 2
    kUTF8 = 3


@struct(order=LittleEndian)
class MHString:
    type: StringType
    length: uint32
    _: padding[8]
    text: F(this.type) >> {
        # Wide characters can be decoded just like this
        StringType.kWString: String(this.length, encoding="utf-16"),
        DEFAULT_OPTION: String(this.length),
    }


# Using this simple class we can simply transform the bytes into
# a dictionary
class PList(Transformer):
    def __init__(self, length) -> None:
        super().__init__(Bytes(length))

    def encode(self, obj: dict, context) -> bytes:
        return plistlib.dumps(obj)

    def decode(self, parsed, context) -> dict:
        return plistlib.loads(parsed)


# NOTE: this is an example of how to apply your own swtich-case logic
# with a context function
def get_object_struct(value: ObjectType, context):
    if value in list(range(1, 15)):
        return MHString

    if value == ObjectType.kPlist:
        return PList(this.total_len - 24)

    # this will be our default option
    return Memory(this.total_len - 24)


@struct(order=LittleEndian)
class MHObjectHeader(MHChunk):  # mhoh
    # The enum can be used here, because we haven't explicitly enabled
    # *strict* mode. The resulting value will be either an integer or
    # the mapped enum value.
    obj_type: ObjectType
    unknown1: uint32
    unknown2: uint32
    data: F(this.obj_type) >> get_object_struct


class RatingType(enum.IntEnum):
    __struct__ = uint8

    kManual = 1
    kAuto = 0x20


@struct(order=LittleEndian)
class MHAlbumItem(MHChunk):  # miah
    child_count: uint32
    unknown: uint32
    item_id: uint64
    unknown_1: uint32
    album_id: uint64
    rating: uint8
    rated_by: RatingType
    _: padding[46]
    objects: MHObjectHeader[this.child_count]


@struct(order=LittleEndian)
class MHAlbumList(MHChunk):  # mlah
    selected: uint32
    _: padding[this.header_len - 16]
    items: MHAlbumItem[this.total_len]


@struct(order=LittleEndian)
class MHImageItem(MHChunk):  # miih
    child_count: uint32
    image_id: uint32
    song_id: uint64
    _: padding[72]
    # Just use an array here
    objects: MHObjectHeader[this.child_count]


@struct(order=LittleEndian)
class MHImageList(MHChunk):  # mlih
    # the last header data is always filled up with zeros
    _: padding[this.header_len - 12]
    items: MHImageItem[this.total_len]


@struct(order=LittleEndian)
class MHTrackItem(MHChunk):  # mtih
    # strings are set at the end
    string_count: uint32
    track_id: uint32
    track_type: uint8
    file_type: uint32
    compiled: boolean
    _unknown: Memory(6)
    play_date: uint32
    length: uint32
    total_time: uint32
    track_number: uint32
    total_tracks: uint32
    year: uint32
    bitrate: uint32
    sample_rate: uint32
    volume_adjustment: int32
    start_time: uint32
    stop_time: uint32
    play_count: uint32
    _1: padding[24]
    disc_num: uint16
    disc_total: uint16
    _2: padding[20]
    persistent_id: uint64
    _3: padding[620]
    strings: MHObjectHeader[this.string_count]


@struct(order=LittleEndian)
class MHTrackList(MHChunk):
    # same list structure
    _: padding[this.header_len - 12]
    tracks: MHTrackItem[this.total_len]


@struct(order=LittleEndian)
class MHSmartPlaylistItem(MHChunk):  # miph
    object_count: uint32
    _: padding[0x20]
    obejcts: MHObjectHeader[this.object_count]


@struct(order=LittleEndian)
class MHSmartPlaylist(MHChunk):  # mlph
    _: padding[0x20]
    items: MHSmartPlaylistItem[this.total_len]


# This enumeration should be subject to future expansion
class MHSubtype(enum.IntEnum):
    __struct__ = uint32

    kFileData = 0x10
    kImageList = 0x0B
    kAlbumList = 0x09
    kTrackList = 0x01
    kPlaylist = 0x02
    kTrackList2 = 0x0D
    kSmartPlaylist = 0x15
    kPlayFilterInfo = 0x17
    kLocation = 0x04
    kRecentPlaylists = 0x0F


# NOTE: all structs in the encrypted part of the database are using little endian
# encoding
@struct(order=LittleEndian)
class MHSectionData(MHChunk):  # msdh
    # We can just use the enum here.
    subtype: MHSubtype
    # all following data must be zeros
    _: padding[this.header_len - 16]
    # The second switch structure, which will be used to implement
    # all subtypes.
    data: F(this.subtype) >> {
        # We can simply use the enum values here as it is an IntEnum
        # and supports comparison with integer objects.
        MHSubtype.kLocation: String(this.total_len - this.header_len),
        MHSubtype.kFileData: MHFileData,
        MHSubtype.kAlbumList: MHAlbumList,
        MHSubtype.kImageList: MHImageList,
        MHSubtype.kTrackList: MHTrackList,
        MHSubtype.kTrackList2: MHTrackList,
        MHSubtype.kSmartPlaylist: MHSmartPlaylist,
        # just copy the raw data
        DEFAULT_OPTION: Memory(this.total_len - this.header_len),
    }


@struct(order=BigEndian)
class ITDB:
    # Actually, the magic must be represented in reverse order, so
    #   hdfm => mhfd, which could be translated to MHFileData
    magic: b"hdfm"
    header_len: uint32
    total_len: uint32
    # The fields below are the same from MHFileData, but with big-endian
    # encoding
    unknown: uint32
    # Prefixed wouldn't help here, as the string is fixed. We could add
    # a padding to the prefixed class.
    version: Prefixed(uint8, encoding="utf-8")
    _: padding[31 - lenof(this.version)]
    chunk_count: uint32
    # This id is important to iTunes
    library_id: uint64
    # the rest of the data will be copied
    pad: Memory(this.header_len - 60)
    # Chain + greedy parsing
    sections: ITDBEncrypted & MHSectionData[...]


# TODO: implement other chunks
if __name__ == "__main__":
    import sys

    KEY = sys.argv[2].encode()
    obj = unpack_file(ITDB, sys.argv[1])
    print(obj)
    if len(sys.argv) >= 4:
        pack_file(obj, sys.argv[3])



from construct import *

d = Struct(
    "count" / Int32ul,
    "items"
    / Array(
        this.count,
        Struct(
            "num1" / Int8ul,
            "num2" / Int24ul,
            "flags"
            / BitStruct(
                "bool1" / Flag,
                "num4" / BitsInteger(3),
                Padding(4),
            ),
            "fixedarray1" / Array(3, Int8ul),
            "name1" / CString("utf8"),
            "name2" / PascalString(Int8ul, "utf8"),
        ),
    ),
)
d_compiled = d.compile()


# Create the blob with:
# data = d.build(
#     dict(
#         count=1000,
#         items=[
#             dict(
#                 num1=0,
#                 num2=0,
#                 flags=dict(bool1=True, num4=0),
#                 fixedarray1=[0, 0, 0],
#                 name1="...",
#                 name2="...",
#             )
#             for i in range(1000)
#         ],
#     )
# )
# with open("blob", "wb") as f:
#     f.write(data)

if __name__ == "__main__":
    import sys
    import timeit

    try:
        from rich import print
    except ImportError:
        pass

    with open(sys.argv[1], "rb") as fp:
        data = fp.read()

    obj = d.parse(data)
    time = timeit.timeit(lambda: d.parse(data), number=1000) / 1000
    time_compiled = timeit.timeit(lambda: d_compiled.parse(data), number=1000) / 1000
    print("[bold]Parsing measurements:[/]")
    print(f"[bold]default[/]  {time:.10f} sec/call")
    print(f"[bold]compiled[/] {time_compiled:.10f} sec/call\n")

    btime = timeit.timeit(lambda: d.build(obj), number=1000) / 1000
    btime_compiled = timeit.timeit(lambda: d_compiled.build(obj), number=1000) / 1000

    print("[bold]Building measurements:[/]")
    print(f"[bold]default[/]  {btime:.10f} sec/call")
    print(f"[bold]compiled[/] {btime_compiled:.10f} sec/call")



# type: ignore
from caterpillar.py import bitfield, CharFactory, int8, unpack, pack

try:
    from rich import print
except ImportError:
    pass


@bitfield
class Format:
    b1: 1  # inferred uint8 type with a width of one bit
    _: 0  # start new uint8 with 7 unused bits
    b2: (2, CharFactory)  # wraps parsed int to char (string)
    b3: 3 - int8 = 1  # default value is applied -> REVISIT: necessary?
    _1: 3  # unnamed padding to the rest of the byte


print(Format.__struct__.groups)
obj = unpack(Format, b"\x80\x80")
print(obj)
# prints: Format(b1=1, b2='\x02', b3=0, _1=0)
# real_pos:      0123456701234567
# bit_pos:       7654321076543210
#                ---------------- # right to left
# because:       1000000010000000
# breakdown is:  |\_____/\/\_/\_/
#                b1  a   b2 b3 u
# where 'u' marks unused bits (the field name starts with an underscore and
# contains only numbers). The field named '_' is a special definition for
# alignment, missing bits will be added automatically (marked with 'a' in
# breakdown).

print(pack(obj))
# prints: b'\x80\x80'



import os
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--include-dir", action='store_true')

argv = parser.parse_args()
if argv.include_dir:
    print(os.path.join(os.path.dirname(__file__), 'include'))


# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# use this module to import everything Python related
from .registry import TypeConverter, annotation_registry, to_struct
from .byteorder import (
    ByteOrder,
    SysNative,
    Arch,
    system_arch,
    Native,
    BigEndian,
    LittleEndian,
    NetEndian,
    AARCH64,
    ARM,
    ARM64,
    MIPS,
    MIPS64,
    PowerPC,
    PowerPC64,
    RISC_V,
    RISC_V64,
    SPARC,
    SPARC64,
    x86,
    x86_64,
    AMD,
    AMD64,
)
from .context import (
    Context,
    ContextPath,
    ContextLength,
    ConditionContext,
    UnaryExpression,
    BinaryExpression,
    this,
    ctx,
    parent,
    CTX_PARENT,
    CTX_OBJECT,
    CTX_OFFSETS,
    CTX_STREAM,
    CTX_FIELD,
    CTX_VALUE,
    CTX_POS,
    CTX_INDEX,
    CTX_PATH,
    CTX_SEQ,
    CTX_ARCH,
    ExprMixin,
)
from .exception import (
    StructException,
    StreamError,
    DelegationError,
    InvalidValueError,
    Stop,
    OptionError,
    ValidationError,
    DynamicSizeError,
)
from .fields import *  # noqa
from .model import *  # noqa
from .options import (
    Flag,
    F_SEQUENTIAL,
    F_DYNAMIC,
    F_KEEP_POSITION,
    F_OFFSET_OVERRIDE,
    set_field_flags,
    set_struct_flags,
    set_union_flags,
    S_ADD_BYTES,
    S_DISCARD_CONST,
    S_SLOTS,
    S_DISCARD_UNNAMED,
    S_REPLACE_TYPES,
    S_UNION,
    S_EVAL_ANNOTATIONS,
    GLOBAL_BITFIELD_FLAGS,
    GLOBAL_STRUCT_OPTIONS,
    GLOBAL_UNION_OPTIONS,
    GLOBAL_FIELD_FLAGS,
    get_flag,
    get_flags,
    has_flag,
    O_ARRAY_FACTORY,
)
from ._common import WithoutContextVar, iseof, pack_seq, unpack_seq
from .shared import (
    ATTR_ACTION_PACK,
    ATTR_STRUCT,
    Action,
    ATTR_ACTION_UNPACK,
    ATTR_BITS,
    ATTR_BYTEORDER,
    ATTR_SIGNED,
    ATTR_TEMPLATE,
    ATTR_TYPE,
    getstruct,
    hasstruct,
    MODE_PACK,
    MODE_UNPACK,
    typeof,
)

from .model import __all__ as model_all
from .fields import __all__ as fields_all

# pyright: reportUnsupportedDunderAll=false
__all__ = model_all + fields_all + [
    "ExprMixin",
    "WithoutContextVar",
    "AARCH64",
    "AMD",
    "AMD64",
    "ARM",
    "ARM64",
    "Arch",
    "BigEndian",
    "ByteOrder",
    "LittleEndian",
    "MIPS",
    "MIPS64",
    "Native",
    "NetEndian",
    "PowerPC",
    "PowerPC64",
    "RISC_V",
    "RISC_V64",
    "SPARC",
    "SPARC64",
    "SysNative",
    "system_arch",
    "x86",
    "x86_64",
    "BinaryExpression",
    "CTX_ARCH",
    "CTX_FIELD",
    "CTX_INDEX",
    "CTX_OBJECT",
    "CTX_OFFSETS",
    "CTX_PARENT",
    "CTX_PATH",
    "CTX_POS",
    "CTX_SEQ",
    "CTX_STREAM",
    "CTX_VALUE",
    "ConditionContext",
    "Context",
    "ContextLength",
    "ContextPath",
    "UnaryExpression",
    "ctx",
    "parent",
    "this",
    "DelegationError",
    "DynamicSizeError",
    "InvalidValueError",
    "OptionError",
    "Stop",
    "StreamError",
    "StructException",
    "ValidationError",
    "F_DYNAMIC",
    "F_KEEP_POSITION",
    "F_OFFSET_OVERRIDE",
    "F_SEQUENTIAL",
    "Flag",
    "GLOBAL_BITFIELD_FLAGS",
    "GLOBAL_FIELD_FLAGS",
    "GLOBAL_STRUCT_OPTIONS",
    "GLOBAL_UNION_OPTIONS",
    "O_ARRAY_FACTORY",
    "S_ADD_BYTES",
    "S_DISCARD_CONST",
    "S_DISCARD_UNNAMED",
    "S_EVAL_ANNOTATIONS",
    "S_REPLACE_TYPES",
    "S_SLOTS",
    "S_UNION",
    "get_flag",
    "get_flags",
    "has_flag",
    "set_field_flags",
    "set_struct_flags",
    "set_union_flags",
    "TypeConverter",
    "annotation_registry",
    "to_struct",
    "ATTR_ACTION_PACK",
    "ATTR_STRUCT",
    "Action",
    "iseof",
    "pack_seq",
    "unpack_seq",
    "ATTR_ACTION_UNPACK",
    "ATTR_BITS",
    "ATTR_BYTEORDER",
    "ATTR_SIGNED",
    "ATTR_TEMPLATE",
    "ATTR_TYPE",
    "getstruct",
    "hasstruct",
    "MODE_PACK",
    "MODE_UNPACK",
    "typeof",
]



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any

from caterpillar.byteorder import Arch
from caterpillar.exception import DelegationError, StructException
from caterpillar.context import CTX_STREAM, CTX_FIELD, CTX_ARCH, CTX_SEQ
from caterpillar.options import Flag
from caterpillar._common import WithoutContextVar
from caterpillar.shared import getstruct

from ._mixin import FieldStruct
from .common import uint16, uint24, uint32, uint64, uint8
from .common import int16, int32, int64, int24, int8
from .common import UInt, Int


PTR_STRICT = Flag("pointer.strict-mode")


class pointer(int):
    """
    A custom integer subclass representing a pointer to another struct within the stream.

    :ivar Any obj: The associated object, if any.
    """

    def __repr__(self) -> str:
        result = super().__repr__()
        if self.obj is not None:
            result = f"<{type(self.obj).__name__}* {hex(self)}>"
        return result

    def get(self):
        return self.obj


class Pointer(FieldStruct):
    """
    A struct that represents a pointer to another struct within the stream.

    :ivar model: An optional configurable model to parse at the offset position.
    :ivar struct: The configured struct to use.
    """

    __slots__ = ("model", "struct")

    def __init__(
        self,
        struct,
        model=None,
    ) -> None:
        self.struct = struct
        self.model = getstruct(model, model) if model is not None else None

    def __mul__(self, model):
        """
        Create a new Pointer with a specified model.

        :param model: The model to use.
        :return: A new Pointer with the specified model.
        :rtype: Pointer
        """

        return type(self)(self.struct, model)

    def __type__(self):
        """
        Get the type associated with the Pointer.

        :return: The type associated with the Pointer.
        :rtype: type
        """
        return pointer

    def __size__(self, context):
        """
        Get the size of the Pointer struct.

        :param context: The context for size calculation.
        :return: The size of the Pointer struct.
        :rtype: int
        """
        struct = self.struct
        if callable(self.struct):
            struct = self.struct(context)
        return struct.__size__(context)

    def unpack_single(self, context):
        """
        Unpack a single value using the Pointer struct.

        :param context: The context for unpacking.
        :return: The unpacked value.
        """
        struct = self.struct
        if callable(struct):
            struct = self.struct(context)

        stream = context[CTX_STREAM]
        start = stream.tell()
        with WithoutContextVar(context, CTX_SEQ, False):
            value: int = struct.__unpack__(context)
            # cleanup before further parsing
            value = self._clean(value, context)

            if self.model is None:
                return self._create(value, start, None, context)

            offset: int = self._to_offset(value, start, context)
            model_obj = None
            if offset != 0:
                # using an if-statement here reduces time overhead
                # of this branch
                fallback: int = stream.tell()
                try:
                    stream.seek(offset)
                    model_obj = self.model.__unpack__(context)
                except StructException as exc:
                    if context[CTX_FIELD].has_flag(PTR_STRICT):
                        raise DelegationError(
                            "Could not parse model!", context
                        ) from exc
                stream.seek(fallback)
        return self._create(value, start, model_obj, context)

    def pack_single(self, obj, context) -> None:
        """
        Pack a single value using the Pointer struct.

        :param obj: The value to pack.
        :param context: The context for packing.
        """
        struct = self.struct
        if callable(struct):
            struct = self.struct(context)

        with WithoutContextVar(context, CTX_SEQ, False):
            struct.__pack__(int(obj), context)

    def _to_offset(self, value, start: int, context) -> int:
        """
        Convert the pointer value to an offset.

        :param value: The pointer value.
        :param start: The starting position in the stream.
        :param context: The context for the conversion.
        :return: The offset value.
        :rtype: int
        """
        return value

    def _clean(self, value: int, context) -> Any:
        """
        Clean the pointer value.

        :param value: The pointer value.
        :param context: The context for cleaning.
        :return: The cleaned pointer value.
        """
        return value

    def _create(self, value, start: int, model_obj, context):
        """
        Create a new pointer object.

        :param value: The pointer value.
        :param start: The starting position in the stream.
        :param model_obj: The object parsed from the model.
        :param context: The context for creation.
        :return: The created pointer object.
        """
        ptr = pointer(value)
        setattr(ptr, "obj", model_obj)
        return ptr


UNSIGNED_POINTER_TYS = {x.__bits__: x for x in [uint8, uint16, uint24, uint32, uint64]}
SIGNED_POINTER_TYS = {x.__bits__: x for x in [int8, int16, int24, int32, int64]}


def uintptr_fn(context):
    """
    Generator function to decide which struct to use as the pointer type based
    on the current architecture.

    :param _ContextLike context: The input context.
    :return: The struct to use.
    :rtype: _StructLike
    """
    arch: Arch = context._root.get(CTX_ARCH, context[CTX_FIELD].arch)
    return UNSIGNED_POINTER_TYS.get(arch.ptr_size, UInt(arch.ptr_size))


def intptr_fn(context):
    """
    Generator function to decide which struct to use as the pointer type based
    on the current architecture.

    :param _ContextLike context: The input context.
    :return: The struct to use.
    :rtype: _StructLike
    """
    arch: Arch = context._root.get(CTX_ARCH, context[CTX_FIELD].arch)
    return SIGNED_POINTER_TYS.get(arch.ptr_size, Int(arch.ptr_size))


uintptr = Pointer(uintptr_fn)
intptr = Pointer(intptr_fn)


class relative_pointer(pointer):
    """
    A custom integer subclass representing a relative pointer to another struct within the stream.

    :ivar obj: The associated object, if any.
    :ivar base: The base offset.
    :ivar absolute: The absolute offset.
    """

    base: int

    @property
    def absolute(self) -> int:
        """
        Get the absolute offset.

        :return: The absolute offset.
        :rtype: int
        """
        return self + self.base


class RelativePointer(Pointer):
    """
    A struct that represents a relative pointer to another struct within the stream.
    """

    def __type__(self):
        """
        Get the type associated with the RelativePointer.

        :return: The type associated with the RelativePointer.
        :rtype: type
        """
        if self.model is not None:
            return f"relative_pointer[{self.model.__type__().__name__}]"

        return relative_pointer

    def _to_offset(self, value, start: int, context) -> int:
        """
        Convert the relative pointer value to an offset.

        :param value: The relative pointer value.
        :param start: The starting position in the stream.
        :param context: The context for the conversion.
        :return: The offset value.
        :rtype: int
        """
        return start + value

    def _create(self, value, start: int, model_obj, context):
        """
        Create a new relative pointer object.

        :param value: The relative pointer value.
        :param start: The starting position in the stream.
        :param model_obj: The object parsed from the model.
        :param context: The context for creation.
        :return: The created relative pointer object.
        """
        ptr = super()._create(value, start, model_obj, context)
        setattr(ptr, "base", start)
        return ptr


offintptr = RelativePointer(intptr_fn)
offuintptr = RelativePointer(uintptr_fn)



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from .byteorder import (
    AARCH64,
    AMD,
    AMD64,
    ARM,
    ARM64,
    BigEndian,
    LittleEndian,
    PowerPC,
    PowerPC64,
    RISC_V,
    RISC_V64,
    x86,
    x86_64,
)
from .context import ContextPath, ctx, parent, this, ContextLength as lenof
from .context import ctx as C, parent as P, root as G
from .model import (
    bitfield,
    pack,
    pack_file,
    pack_into,
    struct,
    union,
    unpack,
    unpack_file,
    sizeof,
    Sequence as Seq,
)
from .shared import typeof, getstruct, hasstruct
from .registry import to_struct
from . import options as opt
from .fields import Field as F

__all__ = [
    "AARCH64",
    "AMD",
    "AMD64",
    "ARM",
    "ARM64",
    "BigEndian",
    "LittleEndian",
    "PowerPC",
    "PowerPC64",
    "RISC_V",
    "RISC_V64",
    "x86",
    "x86_64",
    "bitfield",
    "pack",
    "pack_file",
    "pack_into",
    "struct",
    "union",
    "unpack",
    "unpack_file",
    "sizeof",
    "Seq",
    "typeof",
    "getstruct",
    "hasstruct",
    "ContextPath",
    "ctx",
    "parent",
    "this",
    "lenof",
    "to_struct",
    "F",
    "opt",
    "C",
    "P",
    "G",
]



"""
Here we define all public functions and types to export into
our global struct. Each API dictionary contains the export
name and its target index. Therefore, one might break the whole
import system if an index is changed.

Make sure to check all standard indices before adding a new
function.
"""

# !! heavily inspired by numpy !!

import os
import pathlib

CAPI_PATH = pathlib.Path(__file__).parent.parent / "capi.dat"
if not CAPI_PATH.exists():
    raise FileNotFoundError(f"File not found: {CAPI_PATH}")

# Reserved for future use
RESERVED_STR = "__reserved__"

CP_TYPES: dict[str, str] = {}
CP_TYPE_API: dict[str, tuple[int, str]] = {}
CP_SRC: list[tuple[str, bool]] = []
CP_FUNC_API: dict[str, int] = {}

global_index = 0

for line in CAPI_PATH.read_text("utf-8").splitlines():
    line = line.strip()
    if line.startswith("#") or not line:
        continue

    def_type, *parts = line.split(":")
    match def_type:
        case "obj":
            # obj:INDEX:NAME:TYPE
            #   Defines a C API object.
            index, name, type_ = parts
            if index != "-":
                index = global_index
                global_index += 1
            CP_TYPE_API[name] = (int(index), type_ if type_ != "-" else "PyTypeObject")

        case "type":
            # type:INDEX:STRUCT_NAME:TYPEDEF_NAME:CAPI_TYPE
            #   Defines a C API type for a C structure. The index is optional and
            #   the CAPI_TYPE will be inferred as PyTypeObject if none set
            index, struct_name, typedef_name, c_api_type = parts

            CP_TYPES[struct_name] = typedef_name
            if index != "-":
                index = global_index
                global_index += 1
                if typedef_name.endswith("Object"):
                    typedef_name = typedef_name[:-6] + "_Type"
                CP_TYPE_API[typedef_name] = (
                    int(index),
                    c_api_type if c_api_type != "-" else "PyTypeObject",
                )

        case "src":
            # src:FILE:IGNORE
            #   Defines the source file (relative to this file) that contains the
            #   function definitions.
            name, *ignored = parts
            if not ignored:
                ignored = False
            CP_SRC.append((name, ignored))

        case "func":
            # func:INDEX:NAME:RETURN_TYPE:REFCOUNT
            #   Defines a C API function. The function must be present within the
            #   source set of this file.
            index, name, *_ = parts
            if index != "-":
                index = global_index
                global_index += 1
                CP_FUNC_API[name] = int(index)

        case _:
            pass


def cp_api_functions() -> dict[str, tuple[str, list[str]]]:
    base_path = os.path.dirname(__file__)

    func_api = {}
    # REVISIT: dirty parsing algorithm
    for f, _ in CP_SRC:
        with open(os.path.join(base_path, "..", f), "r", encoding="utf-8") as c_src:
            while True:
                line = c_src.readline()
                if not line:
                    break

                line = line.strip()
                if line.startswith("/*CpAPI"):
                    # API function tag
                    if not line.endswith("*/"):
                        while not line.endswith("*/"):
                            line = c_src.readline()
                            if not line:
                                break
                    else:
                        line = c_src.readline().strip()

                    rtype = line.strip()
                    name, signature = c_src.readline().strip().split("(", 1)
                    if not signature.endswith(")"):
                        while not signature.endswith(")"):
                            signature += c_src.readline().strip()
                    args = signature[:-1].split(", ")
                    func_api[name] = (rtype, args)

    return func_api



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import warnings

__version__ = "2.6.2"
__release__ = "2.6"
__author__ = "MatrixEditor"


def native_support() -> bool:
    """Return True if native support is available."""
    try:
        # pylint: disable-next=import-outside-toplevel
        from caterpillar import _C

        return True
    except ImportError:
        return False


# Explicitly report deprecation warnings
warnings.filterwarnings("default", module="caterpillar")

__all__ = ["__version__", "__author__", "native_support"]



# type: ignore
from caterpillar.py import struct, pack, unpack, this, uint8, int32, Bytes

try:
    from rich import print
except ImportError:
    pass


@struct
class BaseFormat:
    magic: b"BASE"
    length: uint8


# Extending our struct can be achieved through basic inheritance. All fields
# defined in 'BaseFormat' will precede those in this class. You can verify
# the importation of all fields by using help(Format.__init__).
#
# NOTE: A struct class is also capable of inheriting from non-struct classes,
# which will be ignored during processing.
@struct
class Format(BaseFormat):
    foo: int32
    # Referencing fields from the super class is possible
    data: Bytes(this.length)


obj = Format(length=5, foo=-1, data=bytes([i for i in range(5)]))
data = pack(obj)
print(data)

# as Format extends BaseFormat, we can also just unpack a FormatBase
# object.
print(unpack(BaseFormat, data))
print(unpack(Format, data))



from typing import Any, Callable, Protocol, Type, TypeVar

from caterpillar.abc import (
    _LengthT,
    _StructLike,
    _ContextLambda,
    _SwitchLike,
)
from caterpillar.byteorder import ByteOrder, Arch
from caterpillar.fields._base import Field

_TemplateModelT = TypeVar("_TemplateModelT")

class _ContainsTemplate(Protocol):
    __template__: TemplateInfo

class TemplateTypeVar:
    name: str
    field_kwds: dict[str, Any]
    def __init__(self, name: str, **field_kwds) -> None: ...
    def __getitem__(self, amount: _LengthT) -> TemplateTypeVar: ...
    def __rshift__(self, switch: dict | _SwitchLike) -> TemplateTypeVar: ...
    def __matmul__(self, offset: int | _ContextLambda) -> TemplateTypeVar: ...
    def __set_byteorder__(self, order: ByteOrder) -> TemplateTypeVar: ...
    def __rsub__(self, bits: int | _ContextLambda) -> TemplateTypeVar: ...
    def __floordiv__(self, condition: _ContextLambda | bool) -> TemplateTypeVar: ...
    def to_field(
        self,
        struct: _StructLike | _ContextLambda,
        arch: Arch | None = None,
        default=...,
    ) -> Field: ...

class TemplateInfo:
    required_tys: dict[str, _StructLike]
    positional_tys: dict[str, _StructLike]
    def is_defined(self, name: str) -> bool: ...
    def add_required(self, name: str) -> None: ...
    def add_positional(self, name: str, default: Any = None) -> None: ...

def istemplate(obj: Any) -> bool: ...
def template(
    *args: str | TemplateTypeVar, **kwargs
) -> Callable[[Type[_TemplateModelT]], Type[_TemplateModelT]]: ...
def get_mangled_name(model_ty: type, annotations: dict[str, Any]) -> str: ...
def derive(
    template_ty: Type[_ContainsTemplate],
    *tys_args,
    partial: bool = False,
    name: str | None = None,
    union: bool = False,
    **tys_kwargs,
) -> type: ...



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from caterpillar.abc import _ContextLambda, _ContextLike, _StructLike
from caterpillar.context import ConditionContext
from typing import Any, Dict, List

class ConditionalChain:
    chain: Dict[int, _StructLike]
    conditions: List[_ContextLambda[bool]]
    def __init__(self, struct: _StructLike, condition: _ContextLambda[int]) -> None: ...
    def __type__(self) -> type: ...
    def add(self, struct: _StructLike, func: _ContextLambda[bool]) -> None: ...
    def get_struct(self, context: _ContextLike) -> _StructLike | None: ...
    def __unpack__(self, context: _ContextLike) -> Any: ...
    def __pack__(self, obj: Any, context: _ContextLike) -> None: ...
    def __size__(self, context: _ContextLike) -> int: ...

class If(ConditionContext): ...

class ElseIf(ConditionContext):
    def __enter__(self): ...
    def __exit__(self, *_) -> None: ...

Else: ElseIf



# pylint: disable=protected-access
# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import sys
import inspect
import types
import dataclasses

from caterpillar.abc import _GreedyType
from caterpillar.fields import Field, INVALID_DEFAULT
from caterpillar.model import Struct
from caterpillar.options import S_UNION
from caterpillar.shared import ATTR_TEMPLATE


class TemplateTypeVar:
    """Template type variable.

    These specialised type variables are used within a template definition. They
    support most field operators. Therefore, they can be used in some situations
    where you need to adapt the field type at runtime.

    >>> T = TemplateTypeVar("T")

    Note that there is currently no support for inlined type variables, for example:

    >>> @template(T)
    ... class Foo:
    ...     bar: Enum(Baz, T) # !!! throws an error

    is not possible.
    """

    name: str
    """The bound name of this type variable"""

    field_kwds: dict
    """Arguments that will be passed to the created field instance."""

    def __init__(self, name: str, **field_kwds) -> None:
        self.name = name
        self.field_kwds = field_kwds or {}

    def __repr__(self) -> str:
        # This method is important for documentation and type hints.
        # We try to illustrate the generic type as good as possible.
        count = self.field_kwds.get("amount")
        if not count:
            return f"~{self.name}"

        # display sequence access brackets
        return f"~{self.name}[{count}]"

    # Now we have to implement all special operators defined in FieldMixin
    def __getitem__(self, amount):
        return TemplateTypeVar(self.name, amount=amount, **self.field_kwds)

    def __rshift__(self, switch):
        return TemplateTypeVar(self.name, options=switch, **self.field_kwds)

    def __matmul__(self, offset):
        return TemplateTypeVar(self.name, offset=offset, **self.field_kwds)

    def __set_byteorder__(self, order):
        return TemplateTypeVar(self.name, order=order, **self.field_kwds)

    def __rsub__(self, bits):
        return TemplateTypeVar(self.name, bits=bits, **self.field_kwds)

    # @scheduled_for_removal
    def __floordiv__(self, condition):
        return TemplateTypeVar(self.name, condition=condition, **self.field_kwds)

    def to_field(
        self,
        struct,
        arch=None,
        default=INVALID_DEFAULT,
    ) -> Field:
        # REVISIT: what about flags?
        return Field(struct, arch=arch, default=default, **self.field_kwds)


def get_caller_module(frame: int = 1) -> str:
    try:
        # Direct call to retrieve the module name
        return sys._getframemodulename(frame) or "__main__"
    except AttributeError:
        # Sometimes the call fails and we have to navigate manually
        try:
            return sys._getframe(frame).f_globals.get("__name__", "__main__")
        except (AttributeError, ValueError):
            # Use inspect module instead
            pass
    try:
        stack = inspect.stack()[frame]
        return stack.frame.f_globals.get("__name__", "__main__")
    except (AttributeError, ValueError) as e:
        raise ModuleNotFoundError("Could not load module from caller!") from e


@dataclasses.dataclass
class TemplateInfo:
    required_tys: dict
    positional_tys: dict

    def is_defined(self, name: str) -> bool:
        return name in list(self.required_tys) + list(self.positional_tys)

    def add_required(self, name: str) -> None:
        if self.is_defined(name):
            raise ValueError(f"Typename {name!r} already defined!")
        self.required_tys[name] = None

    def add_positional(self, name: str, default=None) -> None:
        if self.is_defined(name):
            raise ValueError(f"Typename {name!r} already defined!")
        self.positional_tys[name] = default


def istemplate(obj) -> bool:
    """Return true if the object is a template."""
    return hasattr(obj, ATTR_TEMPLATE)


def template(*args, **kwargs):
    """
    Defines required template type variables if necessary and prepares
    template class definition.

    :return: a wrapper function that will be called with the class instance
    :rtype: Callable[[type], type]
    """
    if len(args) == 0 and len(kwargs) == 0:
        raise ValueError("Template class needs at least one template type var")

    info = TemplateInfo({}, {})
    module = sys.modules[get_caller_module(frame=2)]
    disposable = []
    for value in args:
        var: TemplateTypeVar = None
        match value:
            case str():
                info.add_required(value)
                var = TemplateTypeVar(value)
            case TemplateTypeVar():
                info.add_required(value.name)
                var = value
            case _:
                raise TypeError(f"Invalid typename type: {value!r}")

        # the type variable will be set globally (at least for
        # class creation)
        if not hasattr(module, var.name):
            setattr(module, var.name, var)
            disposable.append(var.name)

    for name, value in kwargs:
        # ellipsis indicates no default value
        if isinstance(value, _GreedyType):
            value = None

        info.add_positional(name, value)
        if not hasattr(module, name):
            setattr(module, name, TemplateTypeVar(name))
            disposable.append(name)

    # the class will get special attributes that identify it as
    # a template class
    def create_template_class(cls) -> type:
        for name in disposable:
            # Only temporary template vars will be removed
            delattr(module, name)
        setattr(cls, ATTR_TEMPLATE, info)
        return cls

    return create_template_class


def get_mangled_name(model_ty: type, annotations: dict) -> str:
    ty_name = model_ty.__name__
    parts = []
    for name, value in annotations.items():
        parts.append(str(hash(f"{name}{value!r}")))

    hex_name = format(hash("".join(parts)), "X").replace("-", "_")
    return f"_{hex_name}{ty_name}"


def derive(
    template_ty, *tys_args, partial=False, name=None, union=False, **tys_kwargs
) -> type:
    """Creates a new struct class based on the given template class.

    :param template_ty: the template class
    :type template_ty: type
    :param partial: whether the resulting class is also a template, defaults to False
    :type partial: bool, optional
    :param name: the new class name, :code:`...` infers the outer variable name, defaults to None
    :type name: str | Ellipsis, optional
    :return: the derived type
    :rtype: type
    """
    if len(tys_args) == 0 and len(tys_kwargs) == 0:
        raise ValueError(
            (
                "To derive a class from a template class at least one "
                "type argument must be given!"
            )
        )
    if not istemplate(template_ty):
        raise TypeError(f"{template_ty.__name__} is not a template class!")

    info: TemplateInfo = getattr(template_ty, ATTR_TEMPLATE)
    if len(tys_args) > len(info.required_tys):
        raise ValueError(
            f"Expected max. {len(info.required_tys)} positional arguments - got {len(tys_args)}!"
        )

    # update necessary parameters
    required_tys = [x for x in info.required_tys if not info.required_tys[x]]
    for arg_name, value in zip(required_tys[: len(tys_args)], tys_args):
        tys_kwargs[arg_name] = value

    # validate against required type parameters
    if not partial:
        for arg_name, value in info.required_tys.items():
            if arg_name not in tys_kwargs and value is None:
                raise ValueError(f"Missing required type argument: {arg_name!r}")

    if isinstance(name, _GreedyType):
        try:
            # Just a hacky way of getting the variable name if possible
            frame = sys._getframe(1)
            context = inspect.getframeinfo(frame).code_context[0]
            if context.count("=") != 0:
                # definition must be <name> = <...>
                parts = context.split(" = ")
                if len(parts) >= 2:
                    name = parts[0]
        except (AttributeError, KeyError, IndexError, TypeError):
            pass

    if name is None:
        name = get_mangled_name(template_ty, tys_kwargs)

    # IF the target module already stores the new type, then return it
    # directly
    module = get_caller_module(2)
    new_ty = getattr(sys.modules[module], name, None)
    if new_ty is None:
        bases = list(template_ty.__bases__)
        for i, base_ty in enumerate(bases):
            if istemplate(base_ty):
                bases[i] = derive(base_ty, *tys_args, *tys_kwargs)

        new_ty = types.new_class(name, template_ty.__bases__, {})

    # prepare annotations
    annotations = inspect.get_annotations(template_ty)
    replaced = {}
    for name, value in annotations.items():
        if isinstance(value, TemplateTypeVar):
            replacement = tys_kwargs.get(value.name)
            if replacement is None:
                replacement = info.positional_tys.get(
                    value.name, info.positional_tys.get(value.name)
                )

            if replacement is None:
                if partial:
                    # missing types will be replaced later on
                    continue

                raise ValueError(
                    f"Could not find type replacement for {value.name!r} at {name!r}"
                )

            if isinstance(replacement, TemplateTypeVar):
                continue

            annotations[name] = value.to_field(
                replacement,
                default=getattr(template_ty, name, INVALID_DEFAULT),
            )
            replaced[value.name] = replacement

    new_ty.__annotations__ = annotations
    new_ty.__module__ = module
    if not partial:
        new_ty.__struct__ = Struct(new_ty, options={} if not union else {S_UNION})
    else:
        new_info = TemplateInfo(info.required_tys.copy(), info.positional_tys.copy())
        for name, replacement in replaced.items():
            if name in info.required_tys:
                new_info.required_tys[name] = replacement
            elif name in info.positional_tys:
                new_info.positional_tys[name] = replacement
        setattr(new_ty, ATTR_TEMPLATE, new_info)
    return new_ty



# type: ignore
from caterpillar.py import struct, pack, unpack, uint8

try:
    from rich import print
except ImportError:
    pass


@struct
class Format:
    value: uint8
    inner: {"a": uint8, "b": uint8}


obj = Format(value=1, inner={"a": 2, "b": 3})
data = pack(obj)
print(data)
deobj: Format = unpack(Format, data)
print(deobj)
print(type(deobj.inner))  # Context instance



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import sys
import hashlib
import warnings
import zlib

from caterpillar.context import CTX_OBJECT, CTX_STREAM
from caterpillar.exception import StructException, ValidationError
from caterpillar.shared import Action
from caterpillar.fields.hook import IOHook
from ._base import Field
from .common import Bytes, uint32

DEFAULT_DIGEST_PATH = "digest"


class _DigestValue:
    """
    A placeholder class representing a digest value.

    This class is used as a placeholder for digests or checksum values in
    the context of cryptographic algorithms or other hash-like operations.
    """

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"


class Algorithm:
    """
    .. versionadded:: 2.4.0

    A class representing a cryptographic or checksum algorithm.

    This class allows for the creation, updating, and finalization of values using a
    specified algorithm. It abstracts the necessary methods to interact with
    hash algorithms or checksums like SHA256 or CRC32.

    Example usage with SHA256:

    >>> SHA256 = Algorithm(
    ...     create=lambda context: hashlib.sha256(),
    ...     update=lambda algo_obj, data, context: algo_obj.update(data),
    ...     digest=lambda algo_obj, context: algo_obj.digest(),
    ... )

    Example usage with CRC32:

    >>> CRC32 = Algorithm(
    ...     create=lambda context: zlib.crc32(b""),
    ...     update=lambda crc_value, data, context: crc32(data, crc_value),
    ...     digest=lambda crc_value, context: crc_value,
    ... )

    This class can be used to handle both cryptographic algorithms and non-cryptographic checksums.

    :param create: A callable that creates an instance of the algorithm or checksum.
    :type create: Callable[[_ContextLike], Any]
    :param update: A callable that updates the algorithm or checksum with new data.
    :type update: Callable[[Any, bytes, _ContextLike], Any]
    :param digest: A callable that returns the digest or checksum value.
    :type digest: Callable[[Any, _ContextLike], bytes]
    :param name: An optional name for the algorithm.
    :type name: Optional[str]
    """

    __slots__ = ("_create", "_update", "_digest", "name")

    def __init__(self, create=None, update=None, digest=None, name=None) -> None:
        self._create = create
        self._update = update
        self._digest = digest
        self.name = name

    def __repr__(self) -> str:
        """
        Return a string representation of the Algorithm instance.

        The representation includes the name of the algorithm, or '<not set>' if no name is provided.

        :return: The string representation of the algorithm.
        :rtype: str
        """
        return f"<{self.__class__.__name__} name={self.name or '<not set>'!r}>"

    def create(self, context):
        """
        Create an instance of the algorithm or checksum using the provided context.

        This method invokes the `create` callable passed during initialization.

        :param context: The context in which to create the algorithm instance.
        :type context: _ContextLike
        :return: An instance of the algorithm or checksum.
        :rtype: Any
        :raises NotImplementedError: If the `create` method is not implemented for this algorithm.
        """
        if self._create is not None:
            return self._create(context)

        raise NotImplementedError("create() is not implemented for this algorithm")

    def update(self, algo_obj, data: bytes, context):
        """
        Update the algorithm or checksum with the given data.

        This method invokes the `update` callable passed during initialization.

        :param algo_obj: The algorithm instance to update (e.g., a SHA256 object).
        :param data: The data to be used for updating the algorithm.
        :param context: The context in which the update takes place.
        :type algo_obj: Any
        :type data: bytes
        :type context: _ContextLike
        :return: The updated algorithm instance or checksum value.
        :rtype: Any
        :raises NotImplementedError: If the `update` method is not implemented for this algorithm.
        """
        if self._update is not None:
            return self._update(algo_obj, data, context)

        raise NotImplementedError("update() is not implemented for this algorithm")

    def digest(self, algo_obj, context) -> bytes:
        """
        Compute the digest or checksum value from the algorithm instance.

        This method invokes the `digest` callable passed during initialization.

        :param algo_obj: The algorithm instance from which to compute the digest.
        :param context: The context in which the digest operation takes place.
        :type algo_obj: Any
        :type context: _ContextLike
        :return: The computed digest or checksum value.
        :rtype: bytes
        :raises NotImplementedError: If the `digest` method is not implemented for this algorithm.
        """
        if self._digest is not None:
            return self._digest(algo_obj, context)

        raise NotImplementedError("digest() is not implemented for this algorithm")


class Digest:
    """A class to handle the creation, updating, and verification of digests
    using a specified algorithm.

    .. versionadded:: 2.4.0

    .. versionchanged:: 2.4.3

        Python 3.14 is **not** supported, use :class:`DigestField` instead.

    The `Digest` class allows you to integrate hash or checksum algorithms
    into a struct by installing hooks that manage the digest state during
    packing and unpacking. It supports automatic calculation and validation
    of digests during serialization (packing) and deserialization (unpacking).

    Example usage:

    >>> @struct
    ... class Format:
    ...     with Digest(SHA256, Bytes(32), "sha256", verify=True):
    ...         user_data: Bytes(10)
    ...

    This example defines a struct with a `user_data` field and a `sha256` digest
    field. The `Digest` class automatically calculates the SHA256 checksum of
    `user_data` during packing and validates the checksum during unpacking.

    The resulting struct has the following fields:

    >>> Format.__struct__.fields
    [
        Action(Digest.begin),
        Field('user_data', struct=<Bytes>),
        Action(Digest.end),
        Field('sha256', struct=<Bytes>),
        Action(Digest.verify)
    ]

    The `Action` objects install hooks into the current stream for digest calculation
    and validation. When packing an object, the digest will be updated automatically:

    >>> obj = Format(user_data=b"helloworld" * 10)
    Format(user_data=b"helloworld", sha256=<_DigestValue>)

    On packing, the digest is calculated and included in the serialized output:

    >>> pack(obj, Format)
    b"hello world\\xb9M'\\xb9" ...

    If verification is enabled, unpacking will check the validity of the digest:

    >>> unpack(Format, b"invalid data with sha256")
    Traceback (most recent call last):
    ...
    caterpillar.exception.ValidationError: Failed to verify digest of 'sha256'!
    - Expected: <expected_digest_value>
    - Actual: <actual_digest_value>
    Context-Path: <root>.sha256

    :param algorithm: The checksum or cryptographic algorithm used to generate the digest.
    :type algorithm: Algorithm
    :param struct: The struct to which the digest will be attached.
    :type struct: _StructLike
    :param name: An optional name for the digest field (defaults to 'sha256' or another default name).
    :type name: Optional[str]
    :param verify: Whether to enable verification of the digest upon unpacking. Default is False.
    :type verify: bool
    :param path: Optional path to the digest field in the context, used for accessing and storing the digest.
    :type path: Optional[str]
    """

    def __init__(self, algorithm, struct, name=None, verify=False, path=None) -> None:
        if (sys.version_info.major, sys.version_info.minor) >= (3, 14):
            warnings.warn(
                "Python3.14 breaks support for Digest fields. The hash must be calculated "
                "manually until a fix has been released."
            )

        self.algo = algorithm
        self.name = name or DEFAULT_DIGEST_PATH
        if "." in self.name:
            raise ValueError(
                "Digest name must not contain '.' character. Use path instead."
            )

        # IO will be initialized in self.begin
        self._hook = IOHook(io=None, update=self.update)
        self.struct = struct
        self._digest = None
        self._obj = None
        self._verify = verify
        self.path = path or f"{CTX_OBJECT}.{self.name}"

    def _get_annotations(self, frame):
        """
        Retrieve the annotations (i.e., field definitions) from the current frame's local variables.

        :param frame: The current stack frame from which to retrieve annotations.
        :type frame: Any
        :return: A dictionary of annotations for the current struct.
        :rtype: dict[str, Any]
        :raises StructException: If annotations cannot be retrieved from the frame.
        """
        try:
            # This will not work on Python3.14+ but supresses errors
            if "__annotations__" not in frame.f_locals:
                frame.f_locals["__annotations__"] = {}

            return frame.f_locals["__annotations__"]
        except KeyError as exc:
            module = frame.f_locals.get("__module__")
            qualname = frame.f_locals.get("__qualname__")
            msg = f"Could not get annotations in {module} (context={qualname!r})"
            raise StructException(msg) from exc

    def __repr__(self) -> str:
        return f"Digest(algo={self.algo!r}, verify={self._verify!r})"

    def __enter__(self):
        """
        Install the start action for the digest field during struct definition.

        This method is used in a `with` block to set up the digest calculation at the beginning.

        :return: The current Digest instance for further use.
        :rtype: Self
        :raises ValueError: If the start action already exists.
        """
        frame = sys._getframe(1)
        annotations = self._get_annotations(frame)
        # 1. install action to create checksum
        start_action_name = f"{self.name}_start"
        if start_action_name in annotations:
            raise ValueError(
                (
                    f"Digest with start action {self.name!r} already exists! "
                    "Make sure that each digest uses a unique name."
                )
            )

        if self.name in annotations:
            raise ValueError(
                (
                    f"Digest with name {self.name!r} already exists "
                    "before the start action. Make sure that each digest "
                    "uses a unique name."
                )
            )

        annotations[start_action_name] = Action(self.begin, self.begin)
        return self

    def __exit__(self, *_) -> None:
        """
        Install the end action for the digest field and finalize the struct field definitions.

        This method is used to finalize the digest field in a `with` block.

        :raises ValueError: If the end action already exists.
        """
        frame = sys._getframe(1)
        annotations = self._get_annotations(frame)
        # 2. add action to finish checksum
        end_action_name = f"{self.name}_end"
        if end_action_name in annotations:
            raise ValueError(
                (
                    f"Digest with end action {self.name!r} already exists! "
                    "Make sure that each checksum uses a unique name."
                )
            )
        annotations[end_action_name] = Action(self.end_pack, self.end_unpack)
        annotations[self.name] = Field(self.struct)
        frame.f_locals[self.name] = _DigestValue()
        if self._verify:
            annotations[f"{self.name}_verify"] = Action(unpack=self.verfiy)

    def begin(self, context) -> None:
        """
        Initialize the digest calculation at the beginning of packing/unpacking.

        This method sets up the algorithm and prepares the context for checksum calculation.

        :param context: The current context during packing/unpacking.
        :type context: _ContextLike
        """
        self._hook._io = context[CTX_STREAM]
        self._hook.init(context)
        self._obj = self.algo.create(context)

    def end_pack(self, context) -> None:
        """
        Finalize the digest calculation at the end of packing/unpacking.

        This method calculates the digest and stores it in the context for later use.

        :param context: The current context during packing/unpacking.
        :type context: _ContextLike
        """
        self._digest = self.algo.digest(self._obj, context)
        context.__context_setattr__(self.path or self.name, self._digest)
        self._hook.finish(context)

    def end_unpack(self, context) -> None:
        """
        Finalize the digest calculation at the end of unpacking.

        This method retrieves the digest from the context and updates the checksum.

        :param context: The current context during unpacking.
        :type context: _ContextLike
        """
        self._digest = self.algo.digest(self._obj, context)
        self._hook.finish(context)

    def update(self, data: bytes, context) -> None:
        """
        Update the checksum with new data during packing/unpacking.

        This method feeds new data into the algorithm to update the digest.

        :param data: The data to update the checksum with.
        :type data: bytes
        :param context: The current context during packing/unpacking.
        :type context: _ContextLike
        """
        self._obj = self.algo.update(self._obj, data, context) or self._obj

    def verfiy(self, context) -> None:
        """
        Verify the checksum upon unpacking.

        This method checks if the digest calculated during unpacking matches the expected digest.
        If verification fails, a `ValidationError` is raised.

        :param context: The current context during unpacking.
        :type context: _ContextLike
        :raises ValidationError: If the digest verification fails.
        """
        # we assume self._verify is True
        digest = context.__context_getattr__(self.path or self.name)
        if digest != self._digest:
            digest_raw = digest.hex() if isinstance(digest, bytes) else digest
            expected_digest_raw = (
                self._digest.hex() if isinstance(self._digest, bytes) else self._digest
            )
            raise ValidationError(
                (
                    f"Failed to verify digest of {self.name!r} using "
                    f"algorithm {self.algo.name or '<unnamed>'} \n"
                    f" - Expected: {expected_digest_raw!r} \n"
                    f" - Actual: {digest_raw!r}"
                ),
                context,
            )


CTX_DIGEST_OBJ = "_digest_obj"
CTX_DIGEST_HOOK = "_digest_hook"
CTX_DIGEST_ALGO = "_digest_algo"
CTX_DIGEST = "_digest"


class DigestFieldAction:
    """
    .. versionadded:: 2.4.5

    Represents an Action used to initialize a digest computation before struct processing begins.

    This action is intended to be used as a pre-processing step before serializing or parsing
    struct fields that are to be hashed. It injects a digest hook into the IO stream to intercept
    and update digest state as data is read or written.

    The following context variables will be populated based on the given target:
    - ``_digest_obj__<target>``: The current state of the digest computation. (hash object)
    - ``_digest_hook__<target>``: The IO hook associated with the digest computation.
    - ``_digest_algo__<target>``: The algorithm used for the digest computation.

    :param target: The unique name identifying this digest field.
    :type target: str
    :param algorithm: The digest algorithm implementation (must support create/update/digest).
    :type algorithm: Algorithm
    """

    def __init__(self, target: str, algorithm: Algorithm) -> None:
        self.name = target
        self.algo = algorithm
        self._ctx_obj = f"{CTX_DIGEST_OBJ}__{target}"
        self._ctx_hook = f"{CTX_DIGEST_HOOK}__{target}"
        self._ctx_algo = f"{CTX_DIGEST_ALGO}__{target}"

    def update(self, data: bytes, context) -> None:
        """
        Updates the digest object with new data.

        This method is registered with an IO hook so it gets triggered
        automatically during reads/writes to accumulate digest state.
        """
        obj = context[self._ctx_obj]
        new_obj = self.algo.update(obj, data, context)
        context[self._ctx_obj] = new_obj or obj

    def begin(self, context) -> None:
        """
        Initializes the digest algorithm and attaches an IO hook to track data.
        """
        context[self._ctx_obj] = self.algo.create(context)
        context[self._ctx_algo] = self.algo

        hook = IOHook(io=None, update=self.update)
        hook.init(context)
        context[self._ctx_hook] = hook

    # Both packing and unpacking will trigger the same logic
    __action_pack__ = __action_unpack__ = begin


class DigestField:
    """
    .. versionadded:: 2.4.5

    Represents a field in a struct that stores or verifies a digest.

    This struct field computes a digest over all preceding fields and then either:
      - Packs the resulting digest value.
      - Unpacks a stored digest and optionally verifies it.

    Example Usage:

    .. code-block:: python

        @struct
        class Format:
            _hash_begin: DigestField.begin("hash", Sha2_256_Algo)
            user_data: Bytes(10)
            hash: Sha2_256_Field("hash", verify=True) = None


    :param target: The unique name shared with the corresponding :class:`DigestFieldAction`.
    :type target: str
    :param struct:  The struct used to pack/unpack the digest value (e.g., Bytes(32))
    :type struct: _StructLike
    :param verify:  Whether to verify the unpacked digest against the computed one.
    :type verify: bool
    """

    def __init__(self, target: str, struct, verify=False) -> None:
        self.name = target
        self.struct = struct
        self.verify = verify
        self._ctx_obj = f"{CTX_DIGEST_OBJ}__{target}"
        self._ctx_hook = f"{CTX_DIGEST_HOOK}__{target}"
        self._ctx_digest = f"{CTX_DIGEST}__{target}"
        self._ctx_algo = f"{CTX_DIGEST_ALGO}__{target}"

    def __type__(self) -> type:
        """Defines the Python type returned after unpacking (always bytes)."""
        return bytes

    def __size__(self, context) -> int:
        """Returns the size in bytes of the digest field."""
        return self.struct.__size__(context)

    def __pack__(self, obj: None, context) -> None:
        """
        Called during packing. Computes the digest over all previously packed data,
        stores it in the context, finalizes the IO hook, and packs the digest itself.
        """
        digest = context[self._ctx_algo].digest(context[self._ctx_obj], context)
        context.__context_setattr__(self.name, digest)
        context[self._ctx_hook].finish(context)
        self.struct.__pack__(digest, context)

    def __unpack__(self, context):
        """
        Called during unpacking. Computes the digest over all preceding data, reads
        the stored digest, optionally verifies it, and returns the unpacked value.

        :raises ValidationError: If verification is enabled and the digest does not match.
        """
        expected = context[self._ctx_algo].digest(context[self._ctx_obj], context)
        context[self._ctx_digest] = expected
        context[self._ctx_hook].finish(context)

        digest = self.struct.__unpack__(context)
        if self.verify and digest != expected:
            digest_raw = digest.hex() if isinstance(digest, bytes) else digest
            expected_raw = expected.hex() if isinstance(expected, bytes) else expected
            raise ValidationError(
                (
                    f"Failed to verify digest of {self.name!r} using "
                    f"algorithm {context[self._ctx_algo].name or '<unnamed>'} \n"
                    f" - Expected: {expected_raw!r} \n"
                    f" - Actual: {digest_raw!r}"
                ),
                context,
            )

        return digest

    @staticmethod
    def begin(target: str, algo):
        """Factory method to create a DigestFieldAction used at the start of a struct
        to set up hashing for the named digest field.

        :param target: Shared identifier between DigestField and its corresponding action.
        :type target: str
        :param algo:  Digest algorithm (e.g., Sha2_256_Algo)
        :type algo: Algorithm
        :return: An action to initialize hashing in the struct context.
        :rtype: _Action
        """
        return DigestFieldAction(target, algo)


# --- public algorithms ---
def _hash_digest(algo, struct):
    """
    A utility function to create a `Digest` wrapper for a specific hash algorithm and struct.
    The wrapper initializes the digest calculation based on the algorithm provided.

    :param algo: The hash algorithm to be used.
    :type algo: Algorithm
    :param struct: The struct type to use for digest.
    :type struct: _StructLike
    :return: A callable that returns a `Digest` instance.
    :rtype: Callable
    """

    def _wrapper(
        name=None,
        verify=False,
        path=None,
    ) -> Digest:
        return Digest(algo, struct, name, verify, path)

    return _wrapper


def _hash_digest_field(struct):
    def _wrapper(
        name: str,
        verify=False,
    ) -> DigestField:
        return DigestField(target=name, struct=struct, verify=verify)

    return _wrapper


def _hashlib_algo(func):
    """
    Creates an `Algorithm` object from a hash function (e.g., hashlib.sha256).

    :param func: The hash function to be wrapped.
    :type func: Callable
    :return: An `Algorithm` instance that can be used with a Digest.
    :rtype: Algorithm
    """
    return Algorithm(
        create=lambda context: func(),
        update=lambda hash_obj, data, context: hash_obj.update(data),
        digest=lambda hash_obj, context: hash_obj.digest(),
        name=func.__name__,
    )


Crc32_Algo = Algorithm(
    create=lambda context: zlib.crc32(b""),
    update=lambda hash_obj, data, context: zlib.crc32(data, hash_obj),
    digest=lambda hash_obj, context: hash_obj,
    name="crc32",
)
Crc32 = _hash_digest(Crc32_Algo, uint32)
Crc32_Field = _hash_digest_field(uint32)

Adler_Algo = Algorithm(
    create=lambda context: zlib.adler32(b""),
    update=lambda hash_obj, data, context: zlib.adler32(data, hash_obj),
    digest=lambda hash_obj, context: hash_obj,
    name="adler32",
)
Adler = _hash_digest(Adler_Algo, uint32)
Adler_Field = _hash_digest_field(uint32)

try:
    from cryptography.hazmat.primitives import hashes

    def _cryptography_hash_algo(cls):
        """
        Creates an `Algorithm` from a cryptography `hashes.HashAlgorithm` class.

        :param cls: The cryptography hash algorithm class (e.g., `hashes.SHA256`).
        :param args: Additional arguments for initializing the algorithm.
        :param kwargs: Additional keyword arguments for initializing the algorithm.
        :return: An `Algorithm` instance.
        """
        return Algorithm(
            create=lambda context: hashes.Hash(cls()),
            update=lambda hash_obj, data, context: hash_obj.update(data),
            digest=lambda hash_obj, context: hash_obj.finalize(),
            name=getattr(cls, "name"),
        )

    Sha1_Algo = _cryptography_hash_algo(hashes.SHA1)
    Sha2_224_Algo = _cryptography_hash_algo(hashes.SHA224)
    Sha2_256_Algo = _cryptography_hash_algo(hashes.SHA256)
    Sha2_384_Algo = _cryptography_hash_algo(hashes.SHA384)
    Sha2_512_Algo = _cryptography_hash_algo(hashes.SHA512)
    Sha3_224_Algo = _cryptography_hash_algo(hashes.SHA3_224)
    Sha3_256_Algo = _cryptography_hash_algo(hashes.SHA3_256)
    Sha3_384_Algo = _cryptography_hash_algo(hashes.SHA3_384)
    Sha3_512_Algo = _cryptography_hash_algo(hashes.SHA3_512)
    Md5_Algo = _cryptography_hash_algo(hashes.MD5)

    Sha1 = _hash_digest(Sha1_Algo, Bytes(20))
    Sha2_224 = _hash_digest(Sha2_224_Algo, Bytes(28))
    Sha2_256 = _hash_digest(Sha2_256_Algo, Bytes(32))
    Sha2_384 = _hash_digest(Sha2_384_Algo, Bytes(48))
    Sha2_512 = _hash_digest(Sha2_512_Algo, Bytes(64))
    Sha3_224 = _hash_digest(Sha3_224_Algo, Bytes(28))
    Sha3_256 = _hash_digest(Sha3_256_Algo, Bytes(32))
    Sha3_384 = _hash_digest(Sha3_384_Algo, Bytes(48))
    Sha3_512 = _hash_digest(Sha3_512_Algo, Bytes(64))
    Md5 = _hash_digest(Md5_Algo, Bytes(16))

except ImportError:
    Sha1_Algo = _hashlib_algo(hashlib.sha1)
    Sha2_224_Algo = _hashlib_algo(hashlib.sha224)
    Sha2_256_Algo = _hashlib_algo(hashlib.sha256)
    Sha2_384_Algo = _hashlib_algo(hashlib.sha384)
    Sha2_512_Algo = _hashlib_algo(hashlib.sha512)
    Sha3_224_Algo = _hashlib_algo(hashlib.sha3_224)
    Sha3_256_Algo = _hashlib_algo(hashlib.sha3_256)
    Sha3_384_Algo = _hashlib_algo(hashlib.sha3_384)
    Sha3_512_Algo = _hashlib_algo(hashlib.sha3_512)
    Md5_Algo = _hashlib_algo(hashlib.md5)

    Sha1 = _hash_digest(Sha1_Algo, Bytes(32))
    Sha2_224 = _hash_digest(Sha2_224_Algo, Bytes(28))
    Sha2_256 = _hash_digest(Sha2_256_Algo, Bytes(32))
    Sha2_384 = _hash_digest(Sha2_384_Algo, Bytes(48))
    Sha2_512 = _hash_digest(Sha2_512_Algo, Bytes(64))
    Sha3_224 = _hash_digest(Sha3_224_Algo, Bytes(28))
    Sha3_256 = _hash_digest(Sha3_256_Algo, Bytes(32))
    Sha3_384 = _hash_digest(Sha3_384_Algo, Bytes(48))
    Sha3_512 = _hash_digest(Sha3_512_Algo, Bytes(64))
    Md5 = _hash_digest(Md5_Algo, Bytes(16))


Sha1_Field = _hash_digest_field(Bytes(32))
Sha2_224_Field = _hash_digest_field(Bytes(28))
Sha2_256_Field = _hash_digest_field(Bytes(32))
Sha2_384_Field = _hash_digest_field(Bytes(48))
Sha2_512_Field = _hash_digest_field(Bytes(64))
Sha3_224_Field = _hash_digest_field(Bytes(28))
Sha3_256_Field = _hash_digest_field(Bytes(32))
Sha3_384_Field = _hash_digest_field(Bytes(48))
Sha3_512_Field = _hash_digest_field(Bytes(64))
Md5_Field = _hash_digest_field(Bytes(16))


class HMACAlgorithm(Algorithm):
    """
    HMAC (Hash-based Message Authentication Code) algorithm implementation.

    This class wraps an HMAC algorithm using a specified hash function and key.
    """

    def __init__(
        self,
        key,
        algorithm,
    ) -> None:
        super().__init__(name=f"hmac_{algorithm.name}")
        self._key = key
        self._algorithm = algorithm

    def create(self, context):
        """
        Creates an HMAC object with the provided key and algorithm.
        """
        from cryptography.hazmat.primitives import hmac

        key = self._key(context) if callable(self._key) else self._key
        return hmac.HMAC(key, self._algorithm)

    def update(self, algo_obj, data: bytes, context):
        """
        Updates the HMAC object with new data.
        """
        return algo_obj.update(data)

    def digest(self, algo_obj, context):
        """
        Finalizes the HMAC object and returns the computed digest.
        """
        return algo_obj.finalize()


class HMAC(Digest):
    """
    HMAC Digest handler, used to create and verify HMACs based on a provided key and algorithm.
    """

    def __init__(
        self,
        key,
        algorithm,
        name=None,
        verify=False,
        path=None,
    ) -> None:
        super().__init__(
            HMACAlgorithm(key, algorithm),
            Bytes(algorithm.digest_size),
            name,
            verify,
            path,
        )



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from caterpillar.abc import _StructLike
from typing import Any, Callable
from typing_extensions import Self

class TypeConverter:
    target: type
    delegate: Callable[[Any, dict], _StructLike]
    def __init__(
        self,
        target: type | None = None,
        delegate: Callable[[Any, dict], _StructLike] | None = None,
    ) -> None: ...
    def matches(self, annotation: Any) -> bool: ...
    def convert(self, annotation: Any, kwargs: dict) -> _StructLike: ...
    def __call__(self, delegate: Callable[[Any, dict], _StructLike]) -> Self: ...

annotation_registry: list[TypeConverter]

def to_struct(obj: Any, **kwargs) -> _StructLike: ...



# type: ignore
from caterpillar.py import (
    set_struct_flags,
    uintptr,
    CString,
    BigEndian,
    S_REPLACE_TYPES,
    unpack,
    struct,
    x86
)
from caterpillar.fields.pointer import uintptr_fn

try:
    from rich import print
except ImportError:
    pass

set_struct_flags(S_REPLACE_TYPES)


@struct(kw_only=False, order=BigEndian)
class Format:
    address: uintptr * CString(...)


data = b"\x00\x00\x00\x04Hello, World!\x00"
obj = unpack(Format, data, _arch=x86)
print(obj)
print(obj.address.get())



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from io import IOBase
from types import EllipsisType
from typing import (
    Any,
    Callable,
    Optional,
    Protocol,
    TypeVar,
    Union,
    runtime_checkable,
    type_check_only,
)

_IT = TypeVar("_IT", default=Any)
_IT_co = TypeVar("_IT_co", covariant=True, default=Any)
_IT_contra = TypeVar("_IT_contra", contravariant=True, default=Any)
_OT = TypeVar("_OT", default=Any)
_OT_co = TypeVar("_OT_co", covariant=True, default=Any)

_StreamType = IOBase
_StreamFactory = Callable[[], _StreamType]

_GreedyType = EllipsisType
_PrefixedType = slice  # [_StructLike[int, int], NoneType, NoneType]

_LengthT = Union[int, _PrefixedType, _GreedyType, _ContextLambda]

@runtime_checkable
class _ContextLike(Protocol):
    def __init__(self, **kwargs) -> None: ...
    @property
    def _root(self) -> Optional[_ContextLike]: ...
    def __context_getattr__(self, path: str) -> Any: ...
    def __context_setattr__(self, path: str, value: Any) -> None: ...
    def __getitem__(self, key, /) -> Any: ...
    def __setitem__(self, key, value: Any, /) -> None: ...
    def get(self, key: str, default: Any | None = None, /) -> Any: ...

_ContextLambdaReturnT_co = TypeVar(
    "_ContextLambdaReturnT_co", covariant=True, default=Any
)

@runtime_checkable
class _ContextLambda(Protocol[_ContextLambdaReturnT_co]):
    def __call__(self, context: _ContextLike) -> _ContextLambdaReturnT_co: ...

@runtime_checkable
class _StructLike(Protocol[_IT_contra, _OT_co]):
    def __size__(self, context: _ContextLike) -> int: ...
    def __unpack__(self, context: _ContextLike) -> _OT_co: ...
    def __pack__(self, obj: _IT_contra, context: _ContextLike) -> None: ...
    def __type__(self) -> Optional[Union[type, str]]: ...

_StructT = Union[_ContainsStruct[_IT, _OT], _StructLike[_IT, _OT], _ContextLambda]

@runtime_checkable
class _ContainsStruct(Protocol[_IT_contra, _OT]):
    __struct__: _StructLike[_IT_contra, _OT]

@runtime_checkable
class _SupportsActionUnpack(Protocol):
    def __action_pack__(self, context: _ContextLike) -> None:
        pass

@runtime_checkable
class _SupportsActionPack(Protocol):
    def __action_unpack__(self, context: _ContextLike) -> None:
        pass

_ActionLike = Union[_SupportsActionPack, _SupportsActionUnpack]

@runtime_checkable
class _SupportsPack(Protocol[_IT_contra]):
    def __pack__(self, obj: _IT_contra, context: _ContextLike) -> None: ...

@runtime_checkable
class _SupportsSize(Protocol):
    def __size__(self, context: _ContextLike) -> int: ...

@runtime_checkable
class _SupportsType(Protocol):
    def __type__(self) -> type | str | None: ...

@runtime_checkable
class _SupportsUnpack(Protocol[_OT_co]):
    def __unpack__(self, context: _ContextLike) -> _OT_co: ...

_SwitchLambda = Callable[[Any, _ContextLike], _StructLike[_IT, _OT]]
_SwitchLike = Union[
    dict[Any, Union[_ContainsStruct[_IT, _OT], _StructLike[_IT, _OT]]],
    _SwitchLambda[_IT, _OT],
]

@runtime_checkable
class _SupportsBits(Protocol):
    def __bits__(self) -> int: ...

@runtime_checkable
class _ContainsBits(Protocol):
    __bits__: int

@type_check_only
class _ArchLike(Protocol):
    name: str
    ptr_size: int

    def __init__(self, name: str, ptr_size: int) -> None: ...

@type_check_only
class _SupportsSetEndian(Protocol[_OT_co]):
    def __set_byteorder__(self, order: _EndianLike) -> _OT_co: ...

@type_check_only
class _EndianLike(Protocol):
    name: str
    ch: str

    def __add__(self, other: _SupportsSetEndian[_OT]) -> _OT: ...

@type_check_only
class _OptionLike(Protocol):
    name: str
    value: Any

__all__ = [
    "_ContextLike",
    "_ContextLambda",
    "_SupportsActionUnpack",
    "_SupportsActionPack",
    "_SupportsPack",
    "_SupportsUnpack",
    "_SupportsSize",
    "_StructLike",
    "_SupportsType",
    "_ContainsStruct",
    "_SwitchLike",
    "_SupportsBits",
    "_ContainsBits",
]



from collections.abc import Collection, Iterable
from typing import Any, Generic
from typing_extensions import override, Self
from caterpillar.abc import (
    _OT,
    _IT,
    _ArchLike,
    _EndianLike,
    _SupportsSetEndian,
    _ContextLike,
    _LengthT,
    _StructLike,
    _SwitchLike,
    _ContextLambda,
)

BIG_ENDIAN: c_Endian
HOST_ARCH: c_Arch
LITTLE_ENDIAN: c_Endian
NATIVE_ENDIAN: c_Endian

class c_Arch(_ArchLike):
    name: str
    ptr_size: int
    def __init__(self, name: str, ptr_size: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

class c_Endian(_EndianLike):
    ch: str
    name: str
    def __init__(self, name: str, ch: int) -> None: ...
    def __add__(self, other: _SupportsSetEndian[_OT]) -> _OT: ...
    def __eq__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __le__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __radd__(self, other: _SupportsSetEndian[_OT]) -> _OT: ...

class c_Option:
    name: str
    value: Any
    def __init__(self, name: str, value: Any = ...) -> None: ...
    @override
    def __eq__(self, value: object | c_Option, /) -> bool: ...
    @override
    def __hash__(self) -> int: ...

class c_Context(dict[str, Any]):
    def __init__(self, **kwargs: dict[str, Any]) -> None: ...
    def __context_getattr__(self, path: str) -> Any: ...
    def __context_setattr__(self, path: str, value: Any) -> None: ...
    @property
    def _root(self) -> _ContextLike: ...

class LengthInfo:
    length: int
    greedy: bool
    def __init__(self, length: int = ..., greedy: bool = ...) -> None: ...

class Atom(Generic[_IT, _OT]):
    def __pack__(self, obj: _IT, context: _ContextLike) -> None: ...
    def __pack_many__(
        self,
        obj: Iterable[_IT],
        context: _ContextLike,
        lengthinfo: LengthInfo,
    ) -> None: ...
    def __unpack__(self, context: _ContextLike) -> _OT: ...
    def __unpack_many__(
        self,
        context: _ContextLike,
        lengthinfo: LengthInfo,
    ) -> Iterable[_OT]: ...
    def __type__(self) -> type | str | None: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def __bits__(self) -> int: ...

class BuiltinAtom(Atom[_IT, _OT]):
    def __getitem__(self, dim: _LengthT) -> Repeated[_IT, _OT]: ...
    def __rshift__(self, cases: _SwitchLike[Any, Any]) -> Switch: ...

class Repeated(BuiltinAtom[Collection[_IT], Collection[_OT]]):
    atom: _StructLike[_IT, _OT]
    length: _LengthT

    def __init__(self, atom: _StructLike[_IT, _OT], length: _LengthT) -> None: ...
    @override
    def __pack__(self, obj: Collection[_IT], context: _ContextLike) -> None: ...
    @override
    def __unpack__(self, context: _ContextLike) -> Collection[_OT]: ...
    @override
    def __size__(self, context: _ContextLike) -> int: ...
    @override
    def __bits__(self) -> int: ...
    @override
    def __type__(self) -> type[list[_OT]]: ...
    def __set_byteorder__(self, order: _EndianLike) -> Self: ...

class Switch(BuiltinAtom[Any, Any]):
    cases: _SwitchLike[Any, Any]
    atom: _StructLike[Any, Any] | _ContextLambda[Any]

    def __init__(
        self,
        atom: _StructLike[Any, Any] | _ContextLambda[Any],
        cases: _SwitchLike[Any, Any],
    ) -> None: ...
    @override
    def __pack__(self, obj: Any, context: _ContextLike) -> None: ...
    @override
    def __unpack__(self, context: _ContextLike) -> Any: ...
    @override
    def __type__(self) -> type | str | None: ...
    def __set_byteorder__(self, order: _EndianLike) -> Self: ...
    def eval_with_context(
        self,
        value: Any,
        context: _ContextLike,
    ) -> _StructLike[Any, Any]: ...

class Conditional(BuiltinAtom[_IT | None, _OT | None]):
    atom: _StructLike[_IT, _OT]
    condition: _ContextLambda[bool] | bool

    def __init__(
        self,
        atom: _StructLike[_IT, _OT],
        condition: _ContextLambda[bool] | bool,
    ) -> None: ...
    @override
    def __pack__(self, obj: _IT | None, context: _ContextLike) -> None: ...
    @override
    def __unpack__(self, context: _ContextLike) -> _OT | None: ...
    @override
    def __type__(self) -> type | str | None: ...
    def __set_byteorder__(self, order: _EndianLike) -> Self: ...
    def is_enabled(self, context: _ContextLike) -> bool: ...

class AtOffset(BuiltinAtom[_IT, _OT]):
    atom: _StructLike[_IT, _OT]
    offset: int | _ContextLambda[int]
    whence: int
    keep_pos: bool

    def __init__(
        self,
        atom: _StructLike[_IT, _OT],
        offset: int | _ContextLambda[int],
        whence: int = ...,
        keep_pos: bool = ...,
    ) -> None: ...
    @override
    def __pack__(self, obj: _IT, context: _ContextLike) -> None: ...
    @override
    def __unpack__(self, context: _ContextLike) -> _OT: ...
    @override
    def __type__(self) -> type | str | None: ...
    def __set_byteorder__(self, order: _EndianLike) -> Self: ...
    @property
    def is_number(self) -> bool: ...

__all__ = [
    "c_Arch",
    "c_Endian",
    "c_Option",
    "c_Context",
    "BIG_ENDIAN",
    "HOST_ARCH",
    "LITTLE_ENDIAN",
    "NATIVE_ENDIAN",
    "Atom",
    "LengthInfo",
    "Repeated",
    "Switch",
    "Conditional",
    "AtOffset",
    "BuiltinAtom",
]



# type: ignore
from caterpillar.py import uint16, uint32, boolean, union

try:
    from rich import print
except ImportError:
    pass


@union
class Format:
    foo: uint16
    bar: uint32
    baz: boolean


# we don't need any arguments here
obj = Format()
print(obj)

obj.bar = 0xFF00FF00
print(obj)



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import re

from typing_extensions import Self, override

from caterpillar.context import (
    CTX_FIELD,
    CTX_PATH,
    CTX_OBJECT,
    CTX_STREAM,
    CTX_SEQ,
    O_CONTEXT_FACTORY,
)
from caterpillar.byteorder import (
    ByteOrder,
    Arch,
)
from caterpillar.exception import StructException, ValidationError
from caterpillar.options import (
    S_DISCARD_CONST,
    S_DISCARD_UNNAMED,
    S_UNION,
    S_REPLACE_TYPES,
    Flag,
)
from caterpillar.fields import (
    Field,
    INVALID_DEFAULT,
    FieldMixin,
    Const,
)
from caterpillar._common import unpack_seq, pack_seq
from caterpillar.shared import (
    ATTR_ACTION_PACK,
    ATTR_ACTION_UNPACK,
    Action,
    ATTR_BYTEORDER,
)
from caterpillar import registry


class _Member:
    def __init__(self, name, field, include=False, is_action=False):
        self.name = name
        self.field = field
        self.include = include
        self.is_action = is_action
        self.action_unpack = getattr(field, ATTR_ACTION_UNPACK, None)
        self.action_pack = getattr(field, ATTR_ACTION_PACK, None)

    def __repr__(self):
        if self.is_action:
            return f"Action({self.field})"

        return f"Member(name={self.name}, field={self.field}, include={self.include})"

    @override
    def __eq__(self, value: object, /) -> bool:
        return self.field == value or value == self.name


class Sequence(FieldMixin):
    """Default implementation for a sequence of fields.

    The native Python type mapped to this struct class is :code:`dict`. To convert
    a dictionary into a sequence, you can either use the contructor directly or apply
    the type converter for this class:

    >>> to_struct({'a': uint8})
    Sequence(fields=['a'])
    """

    __slots__ = (
        "model",
        "fields",
        "order",
        "arch",
        "options",
        "field_options",
        "_members",
        "is_union",
    )

    def __init__(
        self,
        model,
        order=None,
        arch=None,
        options=None,
        field_options=None,
    ) -> None:
        self.model = model
        self.arch = arch
        self.order = order
        self.options = set(options or [])
        self.field_options = set(field_options or [])

        # these fields will be set or used while processing the model type
        self._members = {}
        self.fields = []
        self.is_union = S_UNION in self.options
        # Process all fields in the model
        self._process_model()

    def __add__(self, sequence: "Sequence") -> Self:
        # We will try to import all fields from the given sequence
        for member in sequence.fields:
            # Ignore duplicates here:
            if member.name in self.fields:
                continue

            self.fields.append(member)
            if member.include and not member.is_action:
                self._members[member.name] = member.field
        return self

    def __sub__(self, sequence: "Sequence") -> Self:
        # By default, we are only removing existing fields.
        for member in sequence.fields:
            self._members.pop(member.name, None)
            if member.name in self.fields:
                self.fields.remove(member)
        return self

    __iadd__ = __add__
    __isub__ = __sub__

    def __type__(self) -> type:
        return dict

    def has_option(self, option: Flag) -> bool:
        """
        Check if the struct has a specific option.

        :param option: The option to check.
        :return: True if the struct has the specified option, else False.
        """
        return option in self.options

    def _included(self, name: str, default, annotation) -> bool:
        """
        Check if a field with the given name should be included.

        :param name: The name of the field.
        :param default: The default value of the field.
        :return: True if the field should be included, else False.
        """
        if self.has_option(S_DISCARD_UNNAMED):
            if re.match(r"^_[0-9]*$", name):
                return False

        if self.has_option(S_DISCARD_CONST):
            if default != INVALID_DEFAULT:
                return False

        return True

    def _set_default(self, name: str, value) -> None:
        pass

    def _process_default(self, name, annotation, had_default=False):
        default = getattr(self.model, name, INVALID_DEFAULT)
        # constant values that are not in the form of fields, structs or types should
        # be wrapped into constant values. For more information, see _process_field
        if isinstance(annotation, Field):
            if annotation.has_condition():
                # Conditional fields always get a default value
                default = None
            annotation = annotation.struct

        match annotation:
            case bytes() | str():
                default = annotation
            # Make it possible to define custom constants
            case Const():
                default = annotation.value

        if self.is_union:
            # Unions will get none as default value for all fields
            default = None

        if default != INVALID_DEFAULT:
            self._set_default(name, default)
        return default

    def _replace_type(self, name: str, type_: type) -> None:
        pass

    def _remove_from_model(self, name: str) -> None:
        pass

    def _process_model(self) -> None:
        """
        Process all fields in the model.
        """
        removables = []
        annotations = self._prepare_fields()
        had_default = False
        for name, annotation in annotations.items():
            if Action.is_action(annotation):
                self.add_action(annotation)
                removables.append(name)
                continue
            # Process each field and its annotation. In addition, fields with a name in
            # the form of '_[0-9]*' will be removed (if enabled)
            default = self._process_default(name, annotation, had_default)
            if default != INVALID_DEFAULT:
                had_default = True
            is_included = self._included(name, default, annotation)
            if not is_included:
                removables.append(name)

            field = self._process_field(name, annotation, default)
            # we call add_field to safely add the created field
            self.add_field(name, field, is_included)
            if self.has_option(S_REPLACE_TYPES):
                # This way we re-annotate all fields in the current model
                self._replace_type(name, field.get_type())

        for name in removables:
            self._remove_from_model(name)

    def _prepare_fields(self):
        return self.model

    def _process_annotation(self, annotation, default, order: ByteOrder, arch: Arch):
        return registry.to_struct(annotation, arch=arch, order=order)

    def _process_field(self, name: str, annotation, default) -> Field:
        """
        Process a field in the model.

        :param name: The name of the field.
        :param annotation: The annotation of the field.
        :param default: The default value of the field.
        :return: The processed field.
        """
        field = None
        struct = None

        # The order and arch of the field can be overridden by the annotation
        # and should not get applied if this seq does not define it.
        order = getattr(annotation, ATTR_BYTEORDER, self.order)
        arch = self.arch
        result = self._process_annotation(annotation, default, order, arch)
        if isinstance(result, Field):
            field = result
        else:
            struct = result

        if struct is not None:
            field = Field(struct, order, arch=arch, default=default)

        if field is None:
            msg = (
                f"Field '{self.model.__name__}.{name}' could not be created, because "
                "the placed annotation does not have a corresponding handler:\n "
                f"type: {type(annotation)},\n annotation: {annotation!r}"
            )
            raise ValidationError(msg)
        field.default = default
        field.order = self.order or field.order
        field.arch = self.arch or field.arch
        field.flags.update(self.field_options)
        # field.flags.update({hash(x): x for x in self.field_options})
        return field

    def add_field(self, name: str, field: Field, included: bool = False) -> None:
        """
        Add a field to the struct.

        :param name: The name of the field.
        :param field: The field to add.
        :param included: True if the field should be included, else False.
        """
        self.fields.append(_Member(name, field, include=included))
        setattr(field, "__name__", name)
        if included:
            self._members[name] = field

    def add_action(self, action) -> None:
        """
        Add an action to the struct.

        :param action: The action to add.
        """
        self.fields.append(_Member(None, action, is_action=True))

    def del_field(self, name: str, field) -> None:
        """
        Remomves a field from this struct.

        :param name: The name of the field.
        :param field: The field to remove.
        """
        self._members.pop(name, None)
        self.fields.remove(field)

    def get_members(self):
        return self._members.copy()

    def __size__(self, context) -> int:
        """
        Get the size of the struct.

        :param context: The context of the struct.
        :return: The size of the struct.
        """
        sizes = []
        base_path = context[CTX_PATH]
        for member in self.fields:
            if member.is_action:
                continue
            field = member.field
            context[CTX_PATH] = f"{base_path}.{member.name}"
            sizes.append(field.__size__(context))

        return max(sizes) if self.is_union else sum(sizes)

    def unpack_one(self, context):
        # At first, we define the object context where the parsed values
        # will be stored
        init_data = O_CONTEXT_FACTORY.value()
        context[CTX_OBJECT] = O_CONTEXT_FACTORY.value(_parent=context)

        base_path = context[CTX_PATH]
        if self.is_union:
            stream = context[CTX_STREAM]
            start = stream.tell()
            max_size = 0

        for member in self.fields:
            if member.is_action:
                if member.action_unpack:
                    member.action_unpack(context)
                continue

            if self.is_union:
                pos = stream.tell()

            # REVISIT: make this a real attribute
            name = member.name
            # The context path has to be changed accordingly
            context[CTX_PATH] = f"{base_path}.{name}"
            result = member.field.__unpack__(context)
            # the object's data shouldn't include removed fields
            context[CTX_OBJECT][name] = result
            if member.include:
                init_data[name] = result

            if self.is_union:
                # This union implementation will cover the max size
                max_size = max(max_size, stream.tell() - pos)
                stream.seek(start)

        obj = init_data
        if self.is_union:
            # Reset the stream position
            stream.seek(start + max_size)
        return obj

    def __unpack__(self, context):
        """
        Unpack the struct from the stream.

        :param stream: The stream to unpack from.
        :param context: The context of the struct.
        :return: The unpacked object.
        """
        base_path = context[CTX_PATH]
        # REVISIT: the name 'this_context' is misleading here
        this_context = O_CONTEXT_FACTORY.value(
            _root=context._root,
            _parent=context,
            _io=context[CTX_STREAM],
            _path=base_path,
        )
        # See __pack__ for more information
        field = context.get("_field")
        if field and context[CTX_SEQ]:
            return unpack_seq(context, self.unpack_one)
        return self.unpack_one(this_context)

    def get_value(self, obj, name: str, field: Field):
        return obj.get(name, None)

    def pack_one(self, obj, context) -> None:
        max_size = 0
        union_field = None
        base_path: str = context[CTX_PATH]

        for member in self.fields:
            if member.is_action:
                if member.action_pack:
                    member.action_pack(context)
                continue

            # The name has to be set (important for current context)
            name = member.name
            field = member.field
            if self.is_union:
                # Union is only applicable for non-dynamic structs
                size: int = field.__size__(context)
                if size > max_size:
                    max_size = size
                    union_field = field
            else:
                # Default behaviour: let the field write its content to the stream.
                context[CTX_PATH] = f"{base_path}.{name}"
                if member.include:
                    value = self.get_value(obj, name, field)
                else:
                    # REVISIT: this line might not be necessary if const fields already
                    # use their internal value.
                    value = field.default if field.default != INVALID_DEFAULT else None
                field.__pack__(value, context)

        if self.is_union:
            if union_field is None:
                raise StructException(
                    "Invalid union config: no fields declared!", context
                )

            name = union_field.__name__
            context[CTX_PATH] = ".".join([base_path, "<value>"])
            # REVISIT: are constant values allowed here? + name validation?
            value = self.get_value(obj, name, union_field)
            union_field.__pack__(value, context)

    def __pack__(self, obj, context) -> None:
        # As structs can be used in field definitions a field will call this struct
        # and could potentially be a sequence. Therefore, we have to check whether we
        # should unpack multiple objects.
        field = context.get(CTX_FIELD)
        if field and context[CTX_SEQ]:
            pack_seq(obj, context, self.pack_one)
        else:
            ctx = O_CONTEXT_FACTORY.value(
                _root=context._root,
                _parent=context,
                _io=context[CTX_STREAM],
                _path=context[CTX_PATH],
                _obj=obj,
            )
            self.pack_one(obj, ctx)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(fields={list(self._members)})"

    def __str__(self) -> str:
        return self.__repr__()


# --- private sequence tyoe converter ---
@registry.TypeConverter(dict)
def _type_converter(annotation, kwargs: dict):
    arch = kwargs.pop("arch", None)
    order = kwargs.pop("order", None)
    return Sequence(model=annotation, order=order, arch=arch)


registry.annotation_registry.insert(0, _type_converter)



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Any
from caterpillar.context import ConditionContext
from caterpillar.exception import ValidationError
from caterpillar.shared import typeof

from ._base import Field


class ConditionalChain:
    """Simplistic conditional chain that represents if-else statements.

    Using this class we can introduce conditional statements into our
    class definition. While this class can't be used in class definitions,
    it may be used outside of them.

    For conditional statements in class definitions, see
    :class:`~caterpillar.fields.conditional.If`,
    :class:`~caterpillar.fields.conditional.ElseIf` or
    :class:`~caterpillar.fields.conditional.Else`.
    """

    __slots__ = "chain", "conditions"

    def __init__(self, struct, condition) -> None:
        self.chain = {}
        self.conditions = []
        self.add(struct, condition)

    def __type__(self):
        target_type = None
        for struct_ty in self.chain.values():
            target_type = target_type | typeof(struct_ty)

        return target_type

    def __repr__(self) -> str:
        annotation = []
        count = len(self.chain)
        for i, entry in enumerate(self.chain.items()):
            func_idx, struct = entry
            func = self.conditions[func_idx]
            if i != count - 1 or func is not None:
                annotation.append(f"{struct.__type__()} if {func!r}")
            else:
                annotation.append(f"else {struct.__type__()}")

        return f"<Chain {', '.join(annotation)}>"

    def add(self, struct, func) -> None:
        idx = len(self.chain)
        self.chain[idx] = struct
        self.conditions.append(func)

    def get_struct(self, context):
        index = 0
        while index < len(self.chain):
            func = self.conditions[index]
            if func is None or func(context):
                return self.chain[index]
            index += 1

    def __unpack__(self, context) -> Any:
        struct = self.get_struct(context)
        return struct.__unpack__(context) if struct else None

    def __pack__(self, obj: Any, context) -> None:
        struct = self.get_struct(context)
        if struct:
            struct.__pack__(obj, context)

    def __size__(self, context) -> int:
        struct = self.get_struct(context)
        return struct.__size__(context) if struct else 0


class If(ConditionContext):
    """If-statement implementation for class definitions.

    .. versionchanged:: 2.4.5

        Python 3.14 is **not** supported.

    .. code-block:: python

        @struct
        class Format:
            a: uint32

            with If(lambda _: GLOBAL_CONSTANT == 33):
                b: uint8

    Note that this class will alter the used fields and cover multiple
    field definitions. In addition, the type annotation will be modified
    to display the condition as well.

    .. note::
        This class is **not** a struct, but a simple context manager.
    """

    # As this class essentially does the same as ConditionContext,
    # we don't have to implement anything. A simple if-statment
    # is using the built-in conditional execution model from the Field
    # class.


class ElseIf(ConditionContext):
    """ElseIf-statement implementation for class definitions.

    .. versionchanged:: 2.4.3

        Python 3.14 is **not** supported.

    .. code-block:: python

        @struct
        class Format:
            a: uint32

            with this.a == 32:
                ...

            with ElseIf(this.a == 34):
                ...
    """

    def __enter__(self):
        self.depth = 3
        super().__enter__()
        self.depth = 2
        # We have to copy all variables here as we want to
        # provide the possibility to re-define some fields.
        self.annotations = self.annotations.copy()
        return self

    def __exit__(self, *_):
        # we have to inspect no only new names but also defined ones
        frame = self.getframe(self.depth, "Could not enter condition context!")
        annotations = frame.f_locals["__annotations__"]

        # inspect defined fields
        for name in set(annotations) & set(self.namelist):
            new_field = annotations[name]
            field = self.annotations[name]
            if field is not new_field:
                # We can assume that the old field is already an instance
                # of Field, otherwise it would have been defined outside
                # the previous condition.
                if not isinstance(field, (Field, ConditionalChain)):
                    msg = (
                        f"The field {name!r} does not appear to be defined in a "
                        "previous condition context. It can't be defined twice!"
                    )
                    raise ValidationError(msg)
                if not isinstance(field, ConditionalChain):
                    # it MUST store a condition
                    if not field.has_condition():
                        msg = (
                            "A field defined outside a condition context can't be "
                            f"overridden in such. (field={name!r})"
                        )
                        raise ValidationError(msg)

                    new_struct = ConditionalChain(field, field.condition)
                    # Reset field's condition
                    field.condition = True
                else:
                    new_struct = field
                new_struct.add(new_field, self.func)
                annotations[name] = new_struct

        # inspect new fields
        self.annotations = annotations
        super().__exit__()


# REVISIT: There is one case where 'ELSE' is not applicable and will cause
# a field to be present at all times. This problem exists if we add fields
# into an else-branch without a previously defined field.
Else = ElseIf(lambda context: True)



# Just import the type as usual after "pip install ."
from example._example import ExampleCAtom

# NOTE: In this example wwe just print the object and don't
# show any additional implementation of packing or unpacking
# data.
obj = ExampleCAtom(5)
print(obj) # <ExampleCAtom 5>


# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import inspect
import dataclasses as dc

from tempfile import TemporaryFile
from io import BytesIO, IOBase
from collections import OrderedDict
from shutil import copyfileobj

from caterpillar.shared import getstruct, hasstruct, ATTR_STRUCT
from caterpillar.abc import _SupportsUnpack, _SupportsSize
from caterpillar.context import O_CONTEXT_FACTORY, CTX_STREAM
from caterpillar.exception import InvalidValueError
from caterpillar.options import (
    S_EVAL_ANNOTATIONS,
    S_UNION,
    S_ADD_BYTES,
    S_SLOTS,
    GLOBAL_STRUCT_OPTIONS,
    GLOBAL_UNION_OPTIONS,
)
from caterpillar.fields import Field, INVALID_DEFAULT
from caterpillar import registry
from caterpillar.shared import MODE_PACK, MODE_UNPACK

from ._base import Sequence


# REVISIT: remove dataclasses dependency
class Struct(Sequence):
    """
    Represents a structured data model for serialization and deserialization.

    :param model: The target class used as the base model.
    :param order: Optional byte order for the fields in the structure.
    :param arch: Global architecture definition (will be inferred on all fields).
    :param options: Additional options specifying what to include in the final class.
    """

    _member_map_: dict
    # An internal field that maps the field names of all class attributes to their
    # corresponding struct fields.

    __slots__ = ("kw_only",)

    def __init__(
        self,
        model: type,
        order=None,
        arch=None,
        options=None,
        field_options=None,
        kw_only=False,
        hook_cls=None,
    ) -> None:
        self.kw_only = kw_only
        options = set(options or [])
        options.update(
            GLOBAL_UNION_OPTIONS if S_UNION in options else GLOBAL_STRUCT_OPTIONS
        )
        super().__init__(
            model=model,
            order=order,
            arch=arch,
            options=options,
            field_options=field_options,
        )
        setattr(self.model, ATTR_STRUCT, self)
        # Add additional options based on the struct's type
        slots = self.has_option(S_SLOTS)
        self.model = dc.dataclass(self.model, kw_only=self.kw_only, slots=slots)

        setattr(self.model, "__class_getitem__", _struct_getitem(self))
        if self.is_union:
            # install a hook
            self._union_hook = (hook_cls or UnionHook)(self)
            setattr(self.model, "__init__", _union_init(self._union_hook))
            setattr(self.model, "__setattr__", _union_setattr(self._union_hook))
        if self.has_option(S_ADD_BYTES):
            setattr(self.model, "__bytes__", _struct_bytes(self))

    def __type__(self):
        return self.model

    def _prepare_fields(self):
        # We will inspect all base classes in reverse order and selectively
        # utilize classes that store a struct instance. Beginning at position
        # -1, concluding at 0, and using a step size of -1:
        for candidate in self.model.__mro__[-1:0:-1]:
            if hasstruct(candidate):
                # Importing all fields instead of the entire struct.
                # The default behavior on importing structs is implemented
                # by the Sequence class.
                self += getstruct(candidate)

        eval_str = self.has_option(S_EVAL_ANNOTATIONS)
        # The why is described in detail here: https://docs.python.org/3/howto/annotations.html
        return inspect.get_annotations(self.model, eval_str=eval_str)

    def _set_default(self, name: str, value) -> None:
        setattr(self.model, name, value)

    def _process_default(self, name, annotation, had_default=False):
        default = super()._process_default(name, annotation, had_default)
        if default is INVALID_DEFAULT and had_default:
            self.kw_only = True
        return default

    def _replace_type(self, name: str, type_: type) -> None:
        self.model.__annotations__[name] = type_

    def _remove_from_model(self, name: str) -> None:
        self.model.__annotations__.pop(name)

    def unpack_one(self, context):
        return self.model(**super().unpack_one(context))

    def get_value(self, obj, name: str, field: Field):
        return getattr(obj, name, None)


# --- private type converter ---
# TODO: documentation
class _StructTypeConverter(registry.TypeConverter):
    def __init__(self) -> None:
        super().__init__()

    def matches(self, annotation) -> bool:
        return isinstance(annotation, type) and getstruct(annotation) is not None

    def convert(self, annotation, kwargs: dict):
        return getstruct(annotation)


registry.annotation_registry.append(_StructTypeConverter())


def _struct_bytes(model: Struct):
    def to_bytes(self) -> bytes:
        return pack(self, model)

    return to_bytes


def _struct_getitem(model: Struct):
    def class_getitem(*args):
        if len(args) == 2:
            _, dim = args
        elif len(args) != 1:
            raise InvalidValueError(f"Expected only one paremeter, got: {args!r}")
        else:
            (dim,) = args

        return Field(model, amount=dim)

    return class_getitem


def _make_struct(
    cls: type,
    options=None,
    order=None,
    arch=None,
    field_options=None,
    kw_only=False,
    hook_cls=None,
) -> type:
    """
    Helper function to create a Struct class.

    :param cls: The target class used as the base model.
    :param options: Additional options specifying what to include in the final class.
    :param order: Optional configuration value for the byte order of a field.
    :param arch: Global architecture definition (will be inferred on all fields).

    :return: The created Struct class.
    """
    _ = Struct(
        cls,
        order=order,
        arch=arch,
        options=options,
        field_options=field_options,
        kw_only=kw_only,
        hook_cls=hook_cls,
    )
    return _.model


def struct(
    cls=None,
    /,
    *,
    options=None,
    order=None,
    arch=None,
    field_options=None,
    kw_only=False,
):
    """
    Decorator to create a Struct class.

    :param cls: The target class used as the base model.
    :param options: Additional options specifying what to include in the final class.
    :param order: Optional configuration value for the byte order of a field.
    :param arch: Global architecture definition (will be inferred on all fields).

    :return: The created Struct class or a wrapper function if cls is not provided.
    """

    def wrap(cls):
        return _make_struct(
            cls,
            order=order,
            arch=arch,
            options=options,
            field_options=field_options,
            kw_only=kw_only,
        )

    if cls is not None:
        return _make_struct(
            cls,
            order=order,
            arch=arch,
            options=options,
            field_options=field_options,
            kw_only=kw_only,
        )

    return wrap


class UnionHook:
    """Implementation of a hook to simulate union types.

    It will hook two methods of the target model type: :code:`__init__` and
    :code:`__setattr__`. Because the constructor calls *setattr* for each
    attribute in the model, we have to intercept it before it gets called
    to set an internal status.

    Internally, these two methods will be translated into :meth:`~UnionHook.__model_init__`
    and :meth:`~UnionHook.__model_setattr__`. Therefore, any class that implements
    these two methods can be used as a union hook.
    """

    struct: Struct
    """The struct reference"""

    max_size: int
    """The static (cached) maximum size of the union"""

    def __init__(self, struct_: Struct) -> None:
        self.struct = struct_
        # Dynamic size is not allowed and will throw an error here
        self.max_size = sizeof(struct_)
        # These attributes are set by default
        self._processing_ = False
        self._model_init_ = struct_.model.__init__

    def __enter__(self) -> None:
        # shortcut for disabling permanent attribute refresh during
        # init and refresh phases
        self._processing_ = True

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        # This variable MUST be reset afterward
        self._processing_ = False

    def __model_init__(self, obj, *args, **kwargs) -> None:
        # since it is possible now, to specify non-kw_only constructors,
        # we have to capture both, args and kwargs
        with self:
            return self._model_init_(obj, *args, **kwargs)

    def __model_setattr__(self, obj, key: str, new_value) -> None:
        # The target attribute will alyaws be set
        object.__setattr__(obj, key, new_value)

        members = self.struct.get_members()
        if self._processing_ or key not in members:
            # Refresh can't be done if:
            #   1) the current instance is alredy being processed
            #   2) the member is not in the internal model
            return

        with self:
            # delegation into method allows for customisation
            self.refresh(obj, key, new_value, members)

    def refresh(self, obj, key: str, new_value, members) -> None:
        # DEFAULT: retrieve the current field and temporarily pack its data
        field = members[key]
        data = pack(new_value, field)

        # Apply a padding to the retrieved data
        stream = BytesIO(data + b"\0" * (self.max_size - len(data)))
        for name, field in members.items():
            # this field shouldn't be updated as it already was
            if name == key:
                continue

            # we can simply use setattr here
            object.__setattr__(obj, name, unpack(field, stream))
            stream.seek(0)


def _union_init(hook):
    # wrapper function to capture the calling instance
    def init(self, *args, **kwargs) -> None:
        return hook.__model_init__(self, *args, **kwargs)

    return init


def _union_setattr(hook):
    # wrapper function to capture the calling instance
    def setattribute(self, key: str, value) -> None:
        hook.__model_setattr__(self, key, value)

    return setattribute


def union(
    cls=None,
    /,
    *,
    options=None,
    order=None,
    arch=None,
    field_options=None,
    kw_only=False,
    hook_cls=None,
):
    """
    Decorator to create a Union class.

    :param cls: The target class used as the base model.
    :param options: Additional options specifying what to include in the final class.
    :param order: Optional configuration value for the byte order of a field.
    :param arch: Global architecture definition (will be inferred on all fields).

    :return: The created Union class or a wrapper function if cls is not provided.
    """
    options = set(list(options or []) + [S_UNION])

    def wrap(cls):
        return _make_struct(
            cls,
            order=order,
            arch=arch,
            options=options,
            field_options=field_options,
            kw_only=kw_only,
            hook_cls=hook_cls,
        )

    if cls is not None:
        return _make_struct(
            cls,
            order=order,
            arch=arch,
            options=options,
            field_options=field_options,
            kw_only=kw_only,
            hook_cls=hook_cls,
        )

    return wrap


def pack(
    obj,
    struct=None,
    /,
    *,
    use_tempfile=False,
    as_field=False,
    **kwds,
) -> bytes:
    """
    Pack an object into a bytes buffer using the specified struct.

    :param obj: The object to pack.
    :param struct: The struct to use for packing.
    :param kwds: Additional keyword arguments to pass to the pack function.

    :return: The packed bytes.
    """
    buffer = BytesIO()
    pack_into(obj, buffer, struct, use_tempfile=use_tempfile, as_field=as_field, **kwds)
    return buffer.getvalue()


def pack_into(
    obj,
    buffer,
    struct=None,
    /,
    *,
    use_tempfile=False,
    as_field=False,
    **kwds,
) -> None:
    """
    Pack an object into the specified buffer using the specified struct.

    This function serializes an object (`obj`) into a given buffer, using
    a struct to define how the object should be packed. Optionally, the
    function can handle temporary files for packing, use a `Field` wrapper
    around the struct, and support additional keyword arguments. The packed
    data is written to the `buffer`.

    Example 1: Packing an object into a bytes buffer

    >>> buffer = BytesIO()
    >>> my_obj = SomeObject()  # Assume SomeObject is a valid object to be packed
    >>> pack_into(my_obj, buffer, struct=SomeStruct())  # Using a specific struct
    >>> buffer.getvalue()
    b"..."

    Example 2: Packing into a file-like stream (e.g., file)

    >>> with open('packed_data.bin', 'wb') as f:
    ...     pack_into(my_obj, f, struct=SomeStruct())  # Pack into a file

    Example 3: Using `as_field` to wrap the struct in a Field before packing

    >>> buffer = BytesIO()
    >>> pack_into(42, buffer, struct=uint8, as_field=True)
    >>> buffer.getvalue()
    b"\\x2a"

    :param obj: The object to pack (could be a plain object or a structure-like object).
    :param buffer: The buffer to pack the object into (a writable stream such as `BytesIO` or a file).
    :param struct: The struct to use for packing. If not specified, will infer from `obj`.
    :param use_tempfile: Whether to use a temporary file for packing (experimental).
    :param as_field: Whether to wrap the struct in a `Field` before packing.
    :param kwds: Additional keyword arguments to pass to the pack function.

    :raises TypeError: If no `struct` is specified and cannot be inferred from the object.
    """
    offsets = OrderedDict()
    # NOTE: we don't have to set _root here because the default root context
    # will be this instance.
    context = O_CONTEXT_FACTORY.value(
        _parent=None,
        _path="<root>",
        _pos=0,
        _offsets=offsets,
        _is_seq=False,
        mode=MODE_PACK,
        **kwds,
    )
    if struct is None:
        struct = getstruct(obj)
    elif as_field:
        struct = Field(struct)
    elif hasstruct(struct):
        struct = getstruct(struct)

    if struct is None:
        raise TypeError("struct must be specified")

    start = 0
    if use_tempfile:
        # NOTE: this implementation is exprimental - use this option with caution.
        with TemporaryFile() as stream:
            context[CTX_STREAM] = stream
            struct.__pack__(obj, context)

            for offset, value in offsets.items():
                stream.seek(start)
                buffer.write(stream.read(offset - start))
                buffer.write(value)
                start = offset
            if len(offsets) == 0:
                stream.seek(0)
                copyfileobj(stream, buffer)

    else:
        # Default implementation: We use an in-memory buffer to store all packed
        # elements and then apply all offset-packed objects.
        stream = BytesIO()
        context[CTX_STREAM] = stream
        struct.__pack__(obj, context)

        content = stream.getbuffer()
        if len(offsets) == 0:
            buffer.write(content)
        else:
            for offset, value in offsets.items():
                buffer.write(content[start:offset])
                buffer.write(value)
                start = offset


def pack_file(
    obj,
    filename: str,
    struct=None,
    /,
    *,
    use_tempfile=False,
    as_field=False,
    **kwds,
) -> None:
    """
    Pack an object into a file using the specified struct.

    :param obj: The object to pack.
    :param filename: The name of the file to write to.
    :param struct: The struct to use for packing.
    :param kwds: Additional keyword arguments to pass to the pack function.

    :return: None
    """
    with open(filename, "w+b") as fp:
        pack_into(obj, fp, struct, use_tempfile=use_tempfile, as_field=as_field, **kwds)


def unpack(
    struct,
    buffer,
    /,
    *,
    as_field=False,
    **kwds,
):
    """
    Unpack an object from a bytes buffer or stream using the specified struct.

    This function takes a `struct` that defines how data should be unpacked,
    a `buffer` (either bytes or a stream) containing the serialized data, and
    returns the unpacked object. If `as_field` is set to True, the `struct` is
    wrapped by a `Field`. Additional keyword arguments are passed to the root
    context as attributes.

    Example:

    >>> buffer = b'\\x00\\x01\\x02\\x03'
    >>> struct = SomeStruct()
    >>> unpack(struct, buffer)
    ...

    :param struct: The struct to use for unpacking (could be a `SupportsUnpack` or `ContainsStruct` object).
    :param buffer: The bytes buffer or stream to unpack from.
    :param as_field: Whether to wrap the struct in a `Field` transformer before unpacking.
    :param kwds: Additional keyword arguments to pass to the unpack function.

    :return: The unpacked object, which is the result of calling `struct.__unpack__(context)`.

    :raises TypeError: If the `struct` is not a valid struct instance.
    """
    # prepare the data stream
    stream = buffer if isinstance(buffer, IOBase) else BytesIO(buffer)
    context = O_CONTEXT_FACTORY.value(
        _path="<root>",
        _parent=None,
        _io=stream,
        **kwds,
        _pos=0,
        _is_seq=False,
        mode=MODE_UNPACK,
    )
    if as_field:
        struct = Field(struct)
    elif hasstruct(struct):
        struct = getstruct(struct)

    if not isinstance(struct, _SupportsUnpack):
        raise TypeError(f"{type(struct).__name__} is not a valid struct instance!")

    return struct.__unpack__(context)


def unpack_file(
    struct,
    filename: str,
    /,
    *,
    as_field=False,
    **kwds,
):
    """
    Unpack an object from a file using the specified struct.

    :param struct: The struct to use for unpacking.
    :param filename: The name of the file to read from.
    :param kwds: Additional keyword arguments to pass to the unpack function.

    :return: The unpacked object.
    """
    with open(filename, "rb") as fp:
        return unpack(struct, fp, as_field=as_field, **kwds)


def sizeof(obj, **kwds) -> int:
    context = O_CONTEXT_FACTORY.value(_parent=None, _path="<root>", **kwds)
    struct_ = obj
    if hasstruct(struct_):
        struct_ = getstruct(struct_)

    if not isinstance(struct_, _SupportsSize):
        raise TypeError(f"{type(struct_).__name__} does not support size calculation!")

    return struct_.__size__(context)



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import types

from typing import Any, Callable, Collection
from caterpillar.abc import _ContextLike, _OT, _IT, _StreamType
from caterpillar.fields._base import Field

class WithoutContextVar:
    context: _ContextLike
    old_value: Any
    value: Any
    name: str
    field: Field
    def __init__(self, context: _ContextLike, name: str, value: Any) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None: ...

def unpack_seq(
    context: _ContextLike, unpack_one: Callable[[_ContextLike], _OT]
) -> Collection[_OT]: ...
def pack_seq(
    seq: Collection[_IT],
    context: _ContextLike,
    pack_one: Callable[[_IT, _ContextLike], None],
) -> None: ...
def iseof(stream: _StreamType) -> bool: ...



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, "API_VERSION", (0, 9)) < (0, 9):
    raise Exception(
        "Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s"
        % (kaitaistruct.__version__)
    )


class Comparison1Kaitai(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.count = self._io.read_u4le()
        self.items = []
        for i in range(self.count):
            self.items.append(Comparison1Kaitai.Item(self._io, self, self._root))

    class Item(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.num1 = self._io.read_u1()
            self.num2_lo = self._io.read_u2le()
            self.num2_hi = self._io.read_u1()
            self.flags = Comparison1Kaitai.Item.Flags(self._io, self, self._root)
            self.fixedarray1 = []
            for i in range(3):
                self.fixedarray1.append(self._io.read_u1())

            self.name1 = (self._io.read_bytes_term(0, False, True, True)).decode(
                "utf-8"
            )
            self.len_name2 = self._io.read_u1()
            self.name2 = (self._io.read_bytes(self.len_name2)).decode("utf-8")

        class Flags(KaitaiStruct):
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._read()

            def _read(self):
                self.bool1 = self._io.read_bits_int_be(1) != 0
                self.num4 = self._io.read_bits_int_be(3)
                self.padding = self._io.read_bits_int_be(4)

        @property
        def num2(self):
            if hasattr(self, "_m_num2"):
                return self._m_num2

            self._m_num2 = (self.num2_hi << 16) | self.num2_lo
            return getattr(self, "_m_num2", None)


if __name__ == "__main__":
    import sys
    import timeit

    try:
        from rich import print
    except ImportError:
        pass

    with open(sys.argv[1], "rb") as fp:
        data = fp.read()

    time = timeit.timeit(lambda: Comparison1Kaitai.from_bytes(data), number=1000) / 1000
    print("[bold]Parsing measurements:[/]")
    print(f"[bold]default[/]  {time:.10f} sec/call")



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from ._base import Field as Field
from .common import Bytes as Bytes, uint32 as uint32
from caterpillar.abc import _ContextLambda, _ContextLike, _StructLike, _ActionLike
from caterpillar.context import CTX_OBJECT as CTX_OBJECT, CTX_STREAM as CTX_STREAM
from caterpillar.exception import (
    StructException as StructException,
    ValidationError as ValidationError,
)
from caterpillar.fields.hook import IOHook as IOHook
from caterpillar.shared import Action as Action
from cryptography.hazmat.primitives import hashes
from typing import Callable, Generic, Optional, Protocol, Type, TypeVar
from typing_extensions import Self

DEFAULT_DIGEST_PATH: str

class _DigestValue: ...

_AlgoObjT = TypeVar("_AlgoObjT")
_AlgoReturnT = TypeVar("_AlgoReturnT", default=bytes)

class Algorithm(Generic[_AlgoObjT, _AlgoReturnT]):
    name: str
    def __init__(
        self,
        create: _ContextLambda[_AlgoObjT] | None = None,
        update: Callable[[_AlgoObjT, bytes, _ContextLike], _AlgoObjT] | None = None,
        digest: Callable[[_AlgoObjT, _ContextLike], _AlgoReturnT] | None = None,
        name: str | None = None,
    ) -> None: ...
    def create(self, context: _ContextLike) -> _AlgoObjT: ...
    def update(
        self, algo_obj: _AlgoObjT, data: bytes, context: _ContextLike
    ) -> _AlgoObjT: ...
    def digest(self, algo_obj: _AlgoObjT, context: _ContextLike) -> _AlgoReturnT: ...

class Digest(Generic[_AlgoObjT, _AlgoReturnT]):
    algo: Algorithm[_AlgoObjT, _AlgoReturnT]
    name: str
    struct: _StructLike[_AlgoReturnT, _AlgoReturnT]
    path: str
    def __init__(
        self,
        algorithm: Algorithm[_AlgoObjT, _AlgoReturnT],
        struct: _StructLike[_AlgoReturnT, _AlgoReturnT],
        name: str | None = None,
        verify: bool = False,
        path: str | None = None,
    ) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *_) -> None: ...
    def begin(self, context: _ContextLike) -> None: ...
    def end_pack(self, context: _ContextLike) -> None: ...
    def end_unpack(self, context: _ContextLike) -> None: ...
    def update(self, data: bytes, context: _ContextLike) -> None: ...
    def verfiy(self, context: _ContextLike) -> None: ...

CTX_DIGEST_OBJ: str
CTX_DIGEST_HOOK: str
CTX_DIGEST_ALGO: str
CTX_DIGEST: str

class DigestFieldAction(Generic[_AlgoObjT, _AlgoReturnT]):
    name: str
    algo: Algorithm[_AlgoObjT, _AlgoReturnT]
    def __init__(
        self, target: str, algorithm: Algorithm[_AlgoObjT, _AlgoReturnT]
    ) -> None: ...
    def update(self, data: bytes, context: _ContextLike) -> None: ...
    def begin(self, context: _ContextLike) -> None: ...
    __action_pack__ = begin
    __action_unpack__ = begin

class DigestField(Generic[_AlgoReturnT], _StructLike[None, _AlgoReturnT]):
    name: str
    struct: _StructLike[_AlgoReturnT, _AlgoReturnT]
    verify: bool
    def __init__(
        self,
        target: str,
        struct: _StructLike[_AlgoReturnT, _AlgoReturnT],
        verify: bool = False,
    ) -> None: ...
    def __type__(self) -> Type[bytes]: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def __pack__(self, obj: None, context: _ContextLike) -> None: ...
    def __unpack__(self, context: _ContextLike) -> _AlgoReturnT: ...
    @staticmethod
    def begin(
        target: str, algo: Algorithm[_AlgoObjT, _AlgoReturnT]
    ) -> DigestFieldAction[_AlgoObjT, _AlgoReturnT]: ...

class _DigestFactory(Protocol[_AlgoReturnT]):
    def __call__(
        self, name: Optional[str] = ..., verify: bool = ..., path: Optional[str] = ...
    ) -> Digest[_AlgoReturnT]: ...

Crc32_Algo: Algorithm[int, int]
Crc32: _DigestFactory[int]
Crc32_Field: DigestField[int]
Adler_Algo: Algorithm[int]
Adler: _DigestFactory[int]
Adler_Field: DigestField[int]
Sha1_Algo: Algorithm[hashes.Hash]
Sha2_224_Algo: Algorithm[hashes.Hash]
Sha2_256_Algo: Algorithm[hashes.Hash]
Sha2_384_Algo: Algorithm[hashes.Hash]
Sha2_512_Algo: Algorithm[hashes.Hash]
Sha3_224_Algo: Algorithm[hashes.Hash]
Sha3_256_Algo: Algorithm[hashes.Hash]
Sha3_384_Algo: Algorithm[hashes.Hash]
Sha3_512_Algo: Algorithm[hashes.Hash]
Md5_Algo: Algorithm[hashes.Hash]
Sha1: _DigestFactory[bytes]
Sha2_224: _DigestFactory[bytes]
Sha2_256: _DigestFactory[bytes]
Sha2_384: _DigestFactory[bytes]
Sha2_512: _DigestFactory[bytes]
Sha3_224: _DigestFactory[bytes]
Sha3_256: _DigestFactory[bytes]
Sha3_384: _DigestFactory[bytes]
Sha3_512: _DigestFactory[bytes]
Md5: _DigestFactory[bytes]

class HMACAlgorithm(Algorithm[hashes.Hash]):
    def __init__(
        self, key: bytes | _ContextLambda[bytes], algorithm: hashes.HashAlgorithm
    ) -> None: ...
    def create(self, context: _ContextLike) -> hashes.Hash: ...
    def update(
        self, algo_obj: hashes.Hash, data: bytes, context: _ContextLike
    ) -> hashes.Hash: ...
    def digest(self, algo_obj: hashes.Hash, context: _ContextLike) -> bytes: ...

class HMAC(Digest[bytes]):
    def __init__(
        self,
        key: bytes | _ContextLambda[bytes],
        algorithm: hashes.HashAlgorithm,
        name: str | None = None,
        verify: bool = False,
        path: str | None = None,
    ) -> None: ...

Sha1_Field: DigestField[bytes]
Sha2_224_Field: DigestField[bytes]
Sha2_256_Field: DigestField[bytes]
Sha2_384_Field: DigestField[bytes]
Sha2_512_Field: DigestField[bytes]
Sha3_224_Field: DigestField[bytes]
Sha3_256_Field: DigestField[bytes]
Sha3_384_Field: DigestField[bytes]
Sha3_512_Field: DigestField[bytes]
Md5_Field: DigestField[bytes]



# type: ignore
from caterpillar.py import struct, uint16, Operator, S_REPLACE_TYPES

# Here, we define a custom operator named 'M' that will multiply
# the second argument by 2.
M = Operator(lambda a, b: a[b*2])

# or directly as function
# @Operator
# def M(a, b):
#     return a[b*2]

@struct(options={S_REPLACE_TYPES})
class Format:
    # __annotations__ should contain typing.List[int] as we've specified the
    # documentation option.
    f1: uint16 /M/ 3

print(Format.__annotations__)

